import logging

from aiohttp import web
from apischema import serialize
from tickit.adapters.http import HttpAdapter
from tickit.adapters.specifications import HttpEndpoint
from tickit.adapters.zmq import ZeroMqPushAdapter

from tickit_devices.eiger.eiger import EigerDevice, get_changed_parameters
from tickit_devices.eiger.eiger_schema import SequenceComplete, construct_value
from tickit_devices.eiger.stream.eiger_stream import EigerStream
from tickit_devices.eiger.stream.eiger_stream_2 import EigerStream2

API_VERSION = "1.8.0"
DETECTOR_API = f"detector/api/{API_VERSION}"
STREAM_API = f"stream/api/{API_VERSION}"
MONITOR_API = "monitor/api/1.8.0"
FILEWRITER_API = "filewriter/api/1.8.0"


def command_404(key: str) -> str:
    return f'error during request: path error: unknown path: "{key}"'


LOGGER = logging.getLogger("EigerAdapter")


class EigerRESTAdapter(HttpAdapter):
    """An Eiger adapter which parses the commands sent to the HTTP server."""

    device: EigerDevice

    def __init__(self, device: EigerDevice) -> None:
        self.device = device

    @HttpEndpoint.get(f"/{DETECTOR_API}" + "/config/{parameter_name}")
    async def get_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting configuration variables from the Eiger.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["parameter_name"]

        if hasattr(self.device.settings, param):
            return web.json_response(construct_value(self.device.settings, param))
        else:
            return web.json_response(status=404)

    @HttpEndpoint.put(f"/{DETECTOR_API}" + "/config/{parameter_name}")
    async def put_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for setting configuration variables for the Eiger.

        Args:
            request (web.Request): The request object that takes the given parameter
            and value.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["parameter_name"]

        response = await request.json()

        if hasattr(self.device.settings, param):
            attr = response["value"]

            LOGGER.debug(f"Changing to {str(attr)} for {str(param)}")

            self.device.settings[param] = attr

            LOGGER.debug("Set " + str(param) + " to " + str(attr))

            changed_parameters = get_changed_parameters(param)

            return web.json_response(serialize(changed_parameters))
        else:
            LOGGER.debug("Eiger has no config variable: " + str(param))
            return web.json_response(status=404)

    @HttpEndpoint.get(
        f"/{DETECTOR_API}" + "/config/threshold/{threshold}/{parameter_name}"
    )
    async def get_threshold_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting threshold configuration from the Eiger.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        threshold = request.match_info["threshold"]
        param = request.match_info["parameter_name"]

        config = self.device.settings.threshold_config
        if threshold in config and hasattr(config[threshold], param):
            return web.json_response(construct_value(config[threshold], param))
        else:
            return web.json_response(status=404)

    @HttpEndpoint.put(
        f"/{DETECTOR_API}" + "/config/threshold/{threshold}/{parameter_name}"
    )
    async def put_threshold_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting threshold configuration from the Eiger.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        threshold = request.match_info["threshold"]
        param = request.match_info["parameter_name"]

        response = await request.json()

        config = self.device.settings.threshold_config
        if threshold in config and hasattr(config[threshold], param):
            attr = response["value"]
            config[threshold][param] = attr

            LOGGER.debug(f"Set threshold/{threshold}{str(param)} to {str(attr)}")

            full_param = f"threshold/{threshold}/{param}"
            changed_parameters = get_changed_parameters(full_param)

            return web.json_response(serialize(changed_parameters))
        else:
            LOGGER.debug("Eiger has no config variable: " + str(param))
            return web.json_response(status=404)

    @HttpEndpoint.get(f"/{DETECTOR_API}" + "/status/{status_param}")
    async def get_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting the status of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["status_param"]

        if hasattr(self.device.status, param):
            data = construct_value(self.device.status, param)

        else:
            return web.json_response(status=404)

        return web.json_response(data)

    @HttpEndpoint.get(f"/{DETECTOR_API}" + "/status/board_000/{status_param}")
    async def get_board_000_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting the status of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        if "th0_temp" in request.message.path:
            data = construct_value(self.device.status, "temperature")
            return web.json_response(data)
        elif "th0_humidity" in request.message.path:
            data = construct_value(self.device.status, "humidity")
            return web.json_response(data)
        return web.json_response(status=404)

    @HttpEndpoint.get(f"/{DETECTOR_API}" + "/status/builder/{status_param}")
    async def get_builder_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting the status of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        return await self.get_status(request)

    @HttpEndpoint.put(f"/{DETECTOR_API}" + "/command/initialize", interrupt=True)
    async def initialize_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'initialize' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        if await request.text() and await request.json():
            return web.json_response(status=404, text=command_404("initialize"))

        await self.device.initialize()

        LOGGER.debug("Initializing Eiger...")
        return web.json_response(serialize(SequenceComplete(1)))

    @HttpEndpoint.put(f"/{DETECTOR_API}" + "/command/arm", interrupt=True)
    async def arm_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'arm' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        if await request.text() and await request.json():
            return web.json_response(status=404, text=command_404("arm"))

        await self.device.arm()

        LOGGER.debug("Arming Eiger...")
        return web.json_response(serialize(SequenceComplete(2)))

    @HttpEndpoint.put(f"/{DETECTOR_API}" + "/command/disarm", interrupt=True)
    async def disarm_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'disarm' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        if await request.text() and await request.json():
            return web.json_response(status=404, text=command_404("disarm"))

        await self.device.disarm()

        LOGGER.debug("Disarming Eiger...")
        return web.json_response(serialize(SequenceComplete(3)))

    @HttpEndpoint.put(f"/{DETECTOR_API}" + "/command/trigger", interrupt=False)
    async def trigger_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'trigger' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        if await request.text() and await request.json():
            # Only expect a parameter in "inte" mode
            if self.device.settings.trigger_mode != "inte":
                return web.json_response(status=404, text=command_404("initialize"))

        LOGGER.debug("Triggering Eiger")
        await self.device.trigger()

        await self.interrupt()
        await self.device.finished_trigger.wait()

        return web.json_response(serialize(SequenceComplete(4)))

    @HttpEndpoint.put(f"/{DETECTOR_API}" + "/command/cancel", interrupt=True)
    async def cancel_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'cancel' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        if await request.text() and await request.json():
            return web.json_response(status=404, text=command_404("cancel"))

        await self.device.cancel()

        LOGGER.debug("Cancelling Eiger...")
        return web.json_response(serialize(SequenceComplete(5)))

    @HttpEndpoint.put(f"/{DETECTOR_API}" + "/command/abort", interrupt=True)
    async def abort_eiger(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for the 'abort' command of the Eiger.

        Args:
            request (web.Request): The request object that takes the request method.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        if await request.text() and await request.json():
            return web.json_response(status=404, text=command_404("abort"))

        await self.device.abort()

        LOGGER.debug("Aborting Eiger...")
        return web.json_response(serialize(SequenceComplete(6)))

    @HttpEndpoint.get(f"/{STREAM_API}" + "/status/{param}")
    async def get_stream_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting status values from the Stream.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]

        if hasattr(self.device.stream_status, param):
            return web.json_response(construct_value(self.device.stream_status, param))
        else:
            return web.json_response(status=404)

    @HttpEndpoint.get(f"/{STREAM_API}" + "/config/{param}")
    async def get_stream_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting config values from the Stream.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]

        if hasattr(self.device.stream_config, param):
            return web.json_response(construct_value(self.device.stream_config, param))
        else:
            return web.json_response(status=404)

    @HttpEndpoint.put(f"/{STREAM_API}" + "/config/{param}")
    async def put_stream_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for setting config values for the Stream.

        Args:
            request (web.Request): The request object that takes the given parameter
            and value.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]

        response = await request.json()

        if hasattr(self.device.stream_config, param):
            attr = response["value"]

            LOGGER.debug(f"Changing to {attr} for {param}")

            self.device.stream_config[param] = attr

            LOGGER.debug("Set " + str(param) + " to " + str(attr))
            return web.json_response([])
        else:
            LOGGER.debug("Eiger has no config variable: " + str(param))
            return web.json_response(status=404)

    @HttpEndpoint.get(f"/{MONITOR_API}" + "/config/{param}")
    async def get_monitor_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting config values from the Monitor.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]

        if hasattr(self.device.monitor_config, param):
            return web.json_response(construct_value(self.device.monitor_config, param))
        else:
            return web.json_response(status=404)

    @HttpEndpoint.put(f"/{MONITOR_API}" + "/config/{param}")
    async def put_monitor_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for setting config values for the Monitor.

        Args:
            request (web.Request): The request object that takes the given parameter
            and value.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]

        response = await request.json()

        if hasattr(self.device.monitor_config, param):
            attr = response["value"]

            LOGGER.debug(f"Changing to {attr} for {param}")

            self.device.monitor_config[param] = attr

            LOGGER.debug("Set " + str(param) + " to " + str(attr))
            return web.json_response([])
        else:
            LOGGER.debug("Eiger has no config variable: " + str(param))
            return web.json_response(status=404)

    @HttpEndpoint.get(f"/{MONITOR_API}" + "/status/{param}")
    async def get_monitor_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting status values from the Monitor.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]

        if hasattr(self.device.monitor_status, param):
            return web.json_response(construct_value(self.device.monitor_status, param))
        else:
            return web.json_response(status=404)

    @HttpEndpoint.get(f"/{FILEWRITER_API}" + "/config/{param}")
    async def get_filewriter_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting config values from the Filewriter.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]

        if hasattr(self.device.filewriter_config, param):
            return web.json_response(
                construct_value(self.device.filewriter_config, param)
            )
        else:
            return web.json_response(status=404)

    @HttpEndpoint.put(f"/{FILEWRITER_API}" + "/config/{param}")
    async def put_filewriter_config(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for setting config values for the Filewriter.

        Args:
            request (web.Request): The request object that takes the given parameter
            and value.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]

        response = await request.json()

        if hasattr(self.device.filewriter_config, param):
            attr = response["value"]

            LOGGER.debug(f"Changing to {attr} for {param}")

            self.device.filewriter_config[param] = attr

            LOGGER.debug("Set " + str(param) + " to " + str(attr))
            return web.json_response([])
        else:
            LOGGER.debug("Eiger has no config variable: " + str(param))
            return web.json_response(status=404)

    @HttpEndpoint.get(f"/{FILEWRITER_API}" + "/status/{param}")
    async def get_filewriter_status(self, request: web.Request) -> web.Response:
        """A HTTP Endpoint for requesting status values from the Filewriter.

        Args:
            request (web.Request): The request object that takes the given parameter.

        Returns:
            web.Response: The response object returned given the result of the HTTP
                request.
        """
        param = request.match_info["param"]

        if hasattr(self.device.filewriter_status, param):
            return web.json_response(
                construct_value(self.device.filewriter_status, param)
            )
        else:
            return web.json_response(status=404)


class EigerZMQAdapter(ZeroMqPushAdapter):
    """An Eiger adapter which parses the data to send along a ZeroMQStream."""

    device: EigerDevice

    def __init__(self, stream: EigerStream | EigerStream2) -> None:
        super().__init__()
        self.stream = stream

    def after_update(self) -> None:
        """Updates IOC values immediately following a device update."""
        if buffered_data := list(self.stream.consume_data()):
            self.add_message_to_stream(buffered_data)



import pytest

from tickit_devices.eiger.stream.stream_status import StreamStatus

# # # # # Eiger StreamStatus Tests # # # # #


@pytest.fixture
def stream_status() -> StreamStatus:
    return StreamStatus()


def test_eiger_stream_status_constructor():
    StreamStatus()


def test_eiger_status_getitem(stream_status):
    assert "ready" == stream_status["state"]["value"]

    with pytest.raises(ValueError):
        stream_status["doesnt_exist"]



from dataclasses import dataclass, field, fields
from typing import Any

from tickit_devices.eiger.eiger_schema import ro_str, ro_uint


def stream_status_keys() -> list[str]:
    # Eiger does not report error as a key
    return ["dropped", "state"]


@dataclass
class StreamStatus:
    """Eiger stream status taken from the API spec."""

    state: str = field(
        default="ready",
        metadata=ro_str(allowed_values=["disabled", "ready", "acquire", "error"]),
    )
    error: list[str] = field(default_factory=lambda: [], metadata=ro_str())
    dropped: int = field(default=0, metadata=ro_uint())

    keys: list[str] = field(default_factory=stream_status_keys)

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        for field_ in fields(self):
            if field_.name == key:
                return {"value": vars(self)[field_.name], "metadata": field_.metadata}
        raise ValueError(f"No field with name {key}")



import pytest

from tickit_devices.eiger.eiger_status import EigerStatus

# # # # # EigerStatus Tests # # # # #


@pytest.fixture
def eiger_status() -> EigerStatus:
    return EigerStatus()


def test_eiger_status_constructor():
    EigerStatus()


def test_eiger_status_getitem(eiger_status):
    assert 24.5 == eiger_status["temperature"]["value"]

    with pytest.raises(ValueError):
        eiger_status["doesnt_exist"]



import asyncio

import pytest
from aioca import caget, caput


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tickit_process", ["examples/configs/multi-ioc.yaml"], indirect=True
)
async def test_femto_system(tickit_process):
    assert (await caget("FEMTO:GAIN_RBV")) == 2.5
    current = await caget("FEMTO:CURRENT")
    assert 100 * 2.5 <= current < 200 * 2.5
    await caput("FEMTO:GAIN", 0.01)
    await asyncio.sleep(0.5)
    assert (await caget("FEMTO:GAIN_RBV")) == 0.01
    current = await caget("FEMTO:CURRENT")
    assert 100 * 0.01 <= current < 200 * 0.01

    async def toggle(expected: bool):
        assert (await caget("PNEUMATIC:FILTER_RBV")) != expected
        await caput("PNEUMATIC:FILTER", expected)
        await asyncio.sleep(0.8)
        assert (await caget("PNEUMATIC:FILTER_RBV")) == expected

    await toggle(True)
    await toggle(False)



import asyncio

import pytest
from aioca import caget, caput
from tickit.core.device import DeviceUpdate
from tickit.core.typedefs import SimTime

from tickit_devices.pneumatic.pneumatic import PneumaticDevice


@pytest.fixture
def pneumatic() -> PneumaticDevice:
    return PneumaticDevice(initial_speed=0.5, initial_state=False)


def test_pneumatic_constructor(pneumatic: PneumaticDevice):
    pass


def test_pneumatic_set_and_get_speed(pneumatic: PneumaticDevice):
    pneumatic.set_speed(3.0)
    assert pneumatic.get_speed() == 3.0


def test_pneumatic_set_and_get_state(pneumatic: PneumaticDevice):
    initial_state = pneumatic.get_state()
    pneumatic.set_state()
    assert pneumatic.target_state is not initial_state


def test_pneumatic_update(pneumatic: PneumaticDevice):
    time = SimTime(0)
    initial_state = pneumatic.get_state()
    pneumatic.set_state()
    device_update: DeviceUpdate = pneumatic.update(time, {})

    assert device_update.outputs["output"] is not initial_state


def test_pneumatic_update_no_state_change(pneumatic: PneumaticDevice):
    time = SimTime(0)
    initial_state = pneumatic.get_state()
    device_update: DeviceUpdate = pneumatic.update(time, {})

    assert device_update.outputs["output"] is initial_state


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tickit_process", ["examples/configs/pneumatic/attns.yaml"], indirect=True
)
async def test_pneumatic_system(tickit_process):
    async def toggle(expected: bool):
        assert (await caget("PNEUMATIC:FILTER_RBV")) != expected
        await caput("PNEUMATIC:FILTER", expected)
        await asyncio.sleep(0.8)
        assert (await caget("PNEUMATIC:FILTER_RBV")) == expected

    await toggle(True)
    await toggle(False)



import pydantic.v1.dataclasses
from tickit.adapters.io import EpicsIo
from tickit.core.adapter import AdapterContainer
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_component import DeviceComponent

from .pneumatic import PneumaticAdapter, PneumaticDevice


@pydantic.v1.dataclasses.dataclass
class Pneumatic(ComponentConfig):
    """Pneumatic simulation with EPICS IOC adapter."""

    initial_speed: float = 2.5
    initial_state: bool = False
    db_file: str = "src/tickit_devices/pneumatic/db_files/filter1.db"
    ioc_name: str = "PNEUMATIC"

    def __call__(self) -> Component:  # noqa: D102
        device = PneumaticDevice(
            initial_speed=self.initial_speed, initial_state=self.initial_state
        )
        adapters = [
            AdapterContainer(
                PneumaticAdapter(device),
                EpicsIo(
                    self.ioc_name,
                    self.db_file,
                ),
            )
        ]
        return DeviceComponent(
            name=self.name,
            device=device,
            adapters=adapters,
        )



from random import uniform
from typing import TypedDict

from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime


class CurrentDevice(Device):
    """The current configured device."""

    class Inputs(TypedDict): ...

    #: A typed mapping containing the current output value
    class Outputs(TypedDict):
        output: float

    def __init__(self, callback_period: int) -> None:
        """Initialise the current device.

        Args:
            callback_period (Optional[int]): The duration in which the device should \
                next be updated. Defaults to int(1e9).
        """
        self.callback_period = SimTime(callback_period)

    def update(self, time: SimTime, inputs) -> DeviceUpdate[Outputs]:
        """Updates the state of the current device.

        Args:
            time (SimTime): The time of the simulation in nanoseconds.
            inputs (State): The state of the input values of the device.

        Returns:
            DeviceUpdate: A container for the Device's outputs and a callback time.
        """
        output = uniform(100, 200)
        print(f"Output! (delta: {time}, inputs: {inputs}, output: {output})")
        return DeviceUpdate(
            self.Outputs(output=output), SimTime(time + self.callback_period)
        )



import pytest
from tickit.core.device import DeviceUpdate
from tickit.core.typedefs import SimTime

from tickit_devices.synchrotron.synchrotron_current import SynchrotronCurrentDevice

INITIAL_CURRENT = 300


@pytest.fixture
def synchrotron_current() -> SynchrotronCurrentDevice:
    return SynchrotronCurrentDevice(initial_current=INITIAL_CURRENT)


def test_synchrotron_current_constructor(synchrotron_current: SynchrotronCurrentDevice):
    pass


def test_synchrotron_current_set_and_get_current(
    synchrotron_current: SynchrotronCurrentDevice,
):
    synchrotron_current.beam_current = INITIAL_CURRENT
    assert synchrotron_current.get_current() == INITIAL_CURRENT


def test_synchrotron_current_lose_current_update(
    synchrotron_current: SynchrotronCurrentDevice,
):
    loss_increment = (270 - INITIAL_CURRENT) / 600
    expected_output = INITIAL_CURRENT + loss_increment
    device_update: DeviceUpdate = synchrotron_current.update(SimTime(0), inputs={})
    assert device_update.outputs["current"] == expected_output


def test_synchrotron_current_topup_fill_update():
    initial_current = 239
    fill_incremenet = (300 - 270) / 15
    expected_output = initial_current + fill_incremenet
    device_update: DeviceUpdate = SynchrotronCurrentDevice(
        initial_current=initial_current
    ).update(SimTime(0), inputs={})
    assert device_update.outputs["current"] == expected_output


def test_synchrotron_current_topup_fill_stops():
    initial_current = 268
    fill_incremenet = (300 - 270) / 15
    expected_output = initial_current + fill_incremenet
    device = SynchrotronCurrentDevice(initial_current=initial_current)
    device_update: DeviceUpdate = device.update(SimTime(0), inputs={})
    assert device_update.outputs["current"] == expected_output

    # mimick a topup for 15 seconds then check the current starts depleting again
    i = 0
    while i < 15:
        i += 1
        new_device_update: DeviceUpdate = device.update(
            SimTime(int(1e9) * i), inputs={}
        )

        assert new_device_update.outputs["current"] > device_update.outputs["current"]
        device_update = new_device_update

    i += 1
    assert (
        device.update(SimTime(int(1e9) * i), inputs={}).outputs["current"]
        < device_update.outputs["current"]
    )


def test_synchrotron_current_call_at(synchrotron_current: SynchrotronCurrentDevice):
    device_update: DeviceUpdate = synchrotron_current.update(SimTime(0), inputs={})
    expected_output = int(1e9)
    assert device_update.call_at == expected_output



from collections.abc import Iterable
from typing import Any

from pydantic.v1 import BaseModel
from zmq import Frame

Json = dict[str, Any]
Sendable = bytes | Frame | memoryview
MultipartMessage = Iterable[Sendable]

DEFAULT_HEADER_TYPE = "dheader-1.0"


class AcquisitionSeriesHeader(BaseModel):
    """Sent before a series of images (and associated headers)."""

    header_detail: str
    series: int
    htype: str = DEFAULT_HEADER_TYPE


class AcquisitionSeriesFooter(BaseModel):
    """Sent at the end of a series of images (and associated headers)."""

    series: int
    htype: str = "dseries_end-1.0"


class AcquisitionDetailsHeader(BaseModel):
    """Describes an additional dataset sent at the beginning of a series.

    Used when header_detail is set to all in AcquisitionSeriesHeader.
    """

    htype: str = DEFAULT_HEADER_TYPE
    shape: tuple[int, int]
    type: str


class ImageHeader(BaseModel):
    """Sent before a detector image blob.

    Metadata about the acquisition operation.
    """

    frame: int
    hash: str
    series: int
    htype: str = "dimage-1.0"


class ImageCharacteristicsHeader(BaseModel):
    """Sent before a detector image blob.

    Metadata about the image.
    """

    encoding: str
    shape: tuple[int, int]
    size: int
    type: str
    htype: str = "dimage_d-1.0"


class ImageConfigHeader(BaseModel):
    """Sent before a detector image blob.

    Describes the metrics on the image acquisition. Time values are ints given in ns.
    """

    real_time: int
    start_time: int
    stop_time: int
    htype: str = "dconfig-1.0"



import pytest

from tickit_devices.eiger.filewriter.filewriter_status import FileWriterStatus

# # # # # Eiger FileWriterStatus Tests # # # # #


@pytest.fixture
def filewriter_status() -> FileWriterStatus:
    return FileWriterStatus()


def test_eiger_filewriter_status_constructor():
    FileWriterStatus()


def test_eiger_status_getitem(filewriter_status):
    assert "ready" == filewriter_status["state"]["value"]

    with pytest.raises(ValueError):
        filewriter_status["doesnt_exist"]



import asyncio
import os
import signal
import sys
from subprocess import PIPE, STDOUT, Popen

import pytest
import pytest_asyncio
from tickit.core.management.event_router import InverseWiring
from tickit.core.management.schedulers.master import MasterScheduler
from tickit.core.state_interfaces.state_interface import get_interface
from tickit.utils.configuration.loading import read_configs

# Prevent pytest from catching exceptions when debugging in vscode so that break on
# exception works correctly (see: https://github.com/pytest-dev/pytest/issues/7409)
if os.getenv("PYTEST_RAISE", "0") == "1":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


# https://docs.pytest.org/en/latest/example/parametrize.html#indirect-parametrization
@pytest.fixture
def tickit_process(request):
    """Subprocess that runs ``tickit all <config_path>``."""
    config_path: str = request.param
    proc = Popen(
        [sys.executable, "-m", "tickit", "all", config_path],
        stdout=PIPE,
        stderr=STDOUT,
        text=True,
    )
    # Wait for IOC to be up
    while True:
        line = proc.stdout.readline()
        if "complete" in line:
            break
    yield proc
    proc.send_signal(signal.SIGINT)
    print(proc.communicate()[0])


@pytest_asyncio.fixture
async def tickit_task(request):
    """Task that runs ``tickit all <config_path>``."""
    config_path: str = request.param
    configs = read_configs(config_path)
    inverse_wiring = InverseWiring.from_component_configs(configs)
    scheduler = MasterScheduler(inverse_wiring, *get_interface("internal"))
    t = asyncio.Task(
        asyncio.wait(
            [
                asyncio.create_task(c().run_forever(*get_interface("internal")))
                for c in configs
            ]
            + [asyncio.create_task(scheduler.run_forever())]
        )
    )
    # TODO: would like to await all_servers_running() here
    await asyncio.sleep(0.5)
    yield t
    tasks = asyncio.tasks.all_tasks()
    for task in tasks:
        task.cancel()
    try:
        await t
    except asyncio.CancelledError:
        pass



from collections.abc import Mapping
from typing import Any
from unittest.mock import ANY

import pytest
from pydantic.v1 import BaseModel

from tickit_devices.eiger.data.dummy_image import Image
from tickit_devices.eiger.data.schema import (
    AcquisitionDetailsHeader,
    AcquisitionSeriesFooter,
    AcquisitionSeriesHeader,
    ImageCharacteristicsHeader,
    ImageConfigHeader,
    ImageHeader,
)
from tickit_devices.eiger.eiger_settings import EigerSettings
from tickit_devices.eiger.stream.eiger_stream import EigerStream


@pytest.fixture
def stream() -> EigerStream:
    return EigerStream()


TEST_SERIES_ID = 1

MINIMAL_HEADER = [
    AcquisitionSeriesHeader(
        header_detail="none",
        series=TEST_SERIES_ID,
    )
]

EIGER_SETTINGS_HEADER = EigerSettings().filtered(
    ["flatfield", "pixelmask", "countrate_correction_table"]
)
X_SIZE = EIGER_SETTINGS_HEADER["x_pixels_in_detector"]
Y_SIZE = EIGER_SETTINGS_HEADER["y_pixels_in_detector"]

BASIC_HEADERS = [
    AcquisitionSeriesHeader(
        header_detail="basic",
        series=TEST_SERIES_ID,
    ),
    EIGER_SETTINGS_HEADER,
]

ALL_HEADERS = [
    AcquisitionSeriesHeader(
        header_detail="all",
        series=TEST_SERIES_ID,
    ),
    EIGER_SETTINGS_HEADER,
    AcquisitionDetailsHeader(
        htype="flatfield-1.0",
        shape=(X_SIZE, Y_SIZE),
        type="float32",
    ),
    ANY,
    AcquisitionDetailsHeader(
        htype="dpixelmask-1.0",
        shape=(X_SIZE, Y_SIZE),
        type="uint32",
    ),
    ANY,
    AcquisitionDetailsHeader(
        htype="dcountrate_table-1.0",
        shape=(2, 1000),
        type="float32",
    ),
    ANY,
]


END_SERIES_FOOTER = [AcquisitionSeriesFooter(series=TEST_SERIES_ID)]


@pytest.mark.parametrize(
    "header_detail,expected_headers",
    [
        ("none", MINIMAL_HEADER),
        ("basic", BASIC_HEADERS),
        ("all", ALL_HEADERS),
    ],
)
def test_begin_series_produces_correct_headers(
    stream: EigerStream,
    header_detail: str,
    expected_headers: list[BaseModel | bytes | Mapping[str, Any]],
) -> None:
    settings = EigerSettings()
    stream.begin_series(settings, TEST_SERIES_ID, header_detail)
    blobs = list(stream.consume_data())

    for a, b in zip(expected_headers, blobs, strict=True):
        assert a == b


@pytest.mark.parametrize("number_of_times", [1, 2])
def test_insert_image_produces_correct_headers_and_blobs(
    stream: EigerStream, number_of_times: int
) -> None:
    for i in range(number_of_times):
        image = Image.create_dummy_image(i, (X_SIZE, Y_SIZE))
        stream.insert_image(image, TEST_SERIES_ID)
        blobs = list(stream.consume_data())
        assert blobs == expected_image_blobs(image)


def test_end_series_produces_correct_headers(
    stream: EigerStream,
) -> None:
    stream.end_series(TEST_SERIES_ID)
    blobs = list(stream.consume_data())
    assert blobs == END_SERIES_FOOTER


def test_data_buffered(stream: EigerStream) -> None:
    settings = EigerSettings()
    stream.begin_series(settings, TEST_SERIES_ID, "all")
    image = Image.create_dummy_image(0, (X_SIZE, Y_SIZE))
    stream.insert_image(image, TEST_SERIES_ID)
    stream.end_series(TEST_SERIES_ID)
    blobs = list(stream.consume_data())

    expected_blobs = ALL_HEADERS + expected_image_blobs(image) + END_SERIES_FOOTER
    for a, b in zip(expected_blobs, blobs, strict=True):
        assert a == b


def expected_image_blobs(image: Image) -> list[bytes | BaseModel]:
    return [
        ImageHeader(
            frame=image.index,
            hash=image.hash,
            series=TEST_SERIES_ID,
        ),
        ImageCharacteristicsHeader(
            encoding=image.encoding,
            shape=image.shape,
            size=len(image.data),
            type=image.dtype,
        ),
        image.data,
        ImageConfigHeader(
            real_time=0,
            start_time=0,
            stop_time=0,
        ),
    ]



import base64
import logging
from collections.abc import Iterable
from pathlib import Path
from queue import Queue
from typing import Any, TypedDict

import cbor2
import numpy as np
from tickit.core.typedefs import SimTime

from tickit_devices.eiger.data.dummy_image import Image
from tickit_devices.eiger.eiger_settings import EigerSettings
from tickit_devices.eiger.stream.stream2 import stream2_tag_decoder

LOGGER = logging.getLogger(__name__)
DATA_PATH = Path(__file__).parent.parent / "data" / "stream2"
STREAM_SETTINGS_MAP = {
    # Direct Mappings
    # TODO: These are ints, but they should be floats
    # https://github.com/DiamondLightSource/tickit-devices/issues/120
    "beam_center_x": "beam_center_x",
    "beam_center_y": "beam_center_y",
    "count_time": "count_time",
    "frame_time": "frame_time",
    "sensor_material": "sensor_material",
    "sensor_thickness": "sensor_thickness",
    # TODO: This is broken because it is now a map of thresholds
    # https://github.com/DiamondLightSource/tickit-devices/issues/121
    # threshold_energy="threshold_energy",
    # Indirect Mappings
    "countrate_correction_enabled": "countrate_correction_applied",
    "detector_description": "description",
    "detector_serial_number": "detector_number",
    "flatfield_enabled": "flatfield_correction_applied",
    "image_size_x": "x_pixels_in_detector",
    "image_size_y": "y_pixels_in_detector",
    "incident_energy": "threshold_energy",
    "incident_wavelength": "wavelength",
    "pixel_mask_enabled": "pixel_mask_applied",
    "pixel_size_x": "x_pixel_size",
    "pixel_size_y": "y_pixel_size",
    "saturation_value": "countrate_correction_count_cutoff",
}
START_ALL_FIELDS = ["flatfield", "pixel_mask", "countrate_correction_lookup_table"]
GONIO_AXES = ["chi", "kappa", "omega", "phi", "two_theta"]


def _load_messages():
    start = image = end = None
    with open(DATA_PATH / "start.cbor", "rb") as f:
        start = cbor2.load(f, tag_hook=stream2_tag_decoder)

    # Populate missing large datasets
    sensor_shape = (start["image_size_y"], start["image_size_x"])
    # we need a base64 encoded array of 4 bit integers, numpy can't provide uint4s
    # we can construct it manually for the trivial zero case
    start["countrate_correction_lookup_table"] = base64.b64encode(65536 // 2 * b"\x00")
    start["flatfield"]["threshold_1"] = base64.b64encode(
        np.prod(sensor_shape) // 2 * b"\x00"  # 2 pixels per byte
    )
    start["pixel_mask"]["threshold_1"] = start["flatfield"]["threshold_1"]  # copy value

    with open(DATA_PATH / "image.cbor", "rb") as f:
        image = cbor2.load(f)
    with open(DATA_PATH / "end.cbor", "rb") as f:
        end = cbor2.load(f)

    return start, image, end


class EigerStream2:
    """Simulation of an Eiger stream."""

    callback_period: SimTime

    _message_buffer: Queue[bytes]

    class Inputs(TypedDict):
        """No inputs."""

    class Outputs(TypedDict):
        """No outputs."""

    def __init__(self, callback_period: int = int(1e9)) -> None:
        """Eiger Stream2 constructor."""
        self.callback_period = SimTime(callback_period)

        self._message_buffer = Queue()

        self._start, self._image, self._end = _load_messages()

    def begin_series(
        self, settings: EigerSettings, series_id: int, header_detail: str
    ) -> None:
        """Send the start message marking the beginning of the acquisition series.

        Args:
            settings: Current detector configuration, a snapshot may be sent with the
                headers.
            series_id: ID for the acquisition series.
            header_detail: Header detail for start message - 'none', 'basic' or 'all'
        """
        if header_detail == "all":
            # Use loaded message in place
            start = self._start
        else:
            # Make a copy with "all" fields removed
            start = {k: v for k, v in self._start.items() if k not in START_ALL_FIELDS}

        # Update message with current state
        # TODO: Check what fields should be updated from current state
        # https://github.com/DiamondLightSource/tickit-devices/issues/122
        start["number_of_images"] = settings.nimages * settings.ntrigger
        for stream_field, setting in STREAM_SETTINGS_MAP.items():
            if stream_field not in start:
                start[stream_field] = getattr(settings, setting)
        for axis in [a for a in GONIO_AXES if a not in start["goniometer"]]:
            # get default values for axes not in start message
            # TODO: Captured cbor start message should have all axes?
            start["goniometer"][axis] = {}
            start["goniometer"][axis]["start"] = float(
                getattr(settings, f"{axis}_start")
            )
            start["goniometer"][axis]["increment"] = float(
                getattr(settings, f"{axis}_increment")
            )

        start["series_id"] = series_id

        self._buffer(cbor_dumps(start))

    def insert_image(self, image: Image, series_id: int) -> None:
        """Send headers and an data blob for a single image.

        Args:
            image: The image with associated metadata
            series_id: ID for the acquisition series.
        """
        self._image["series_id"] = series_id
        self._image["image_id"] = image.index

        self._buffer(cbor_dumps(self._image))

    def end_series(self, series_id: int) -> None:
        """Send footer marking the end of an acquisition series.

        Args:
            series_id: ID of the series to end.
        """
        self._end["series_id"] = series_id
        self._buffer(cbor_dumps(self._end))

    def consume_data(self) -> Iterable[bytes]:
        """Consume all headers and data buffered by other methods.

        Returns:
            Iterable[_Message]: Iterable of headers and data
        """
        while not self._message_buffer.empty():
            message = self._message_buffer.get()
            yield message

    def _buffer(self, message: bytes) -> None:
        self._message_buffer.put_nowait(message)


def cbor_dumps(message: dict[str, Any]) -> bytes:
    """Serialize dictionary to cbor, including headers.

    Args:
        message: Message to be serialized

    """
    return cbor2.dumps(cbor2.CBORTag(55799, message))



import pytest
from aioca import caget

USER_MODE = 4


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tickit_process", ["examples/configs/synchrotron/synchrotron.yaml"], indirect=True
)
async def test_synchrotron_system(tickit_process):
    assert (await caget("CS-CS-MSTAT-01:MODE")) == USER_MODE



import logging
from collections.abc import Mapping
from dataclasses import dataclass, field
from enum import Enum
from functools import partial
from typing import Any, Generic, TypeVar

from apischema import order, serialized
from apischema.fields import with_fields_set
from apischema.metadata import skip
from apischema.serialization import serialize

T = TypeVar("T")

LOGGER = logging.getLogger(__name__)


def field_config(**kwargs) -> Mapping[str, Any]:
    """Helper function to create a typesafe dictionary.

    Helper function to create a typesafe dictionary to be inserted as
    dataclass metadata.

    Args:
        kwargs: Key/value pairs to go into the metadata

    Returns:
        Mapping[str, Any]: A dictionary of {key: value} where all keys are strings
    """
    return dict(**kwargs)


class AccessMode(Enum):
    """Possible access modes for field metadata."""

    READ_ONLY = "r"
    WRITE_ONLY = "w"
    READ_WRITE = "rw"


class ValueType(Enum):
    """Possible value types for field metadata."""

    FLOAT = "float"
    INT = "int"
    UINT = "uint"
    STRING = "string"
    STR_LIST = "string[]"
    BOOL = "bool"
    FLOAT_GRID = "float[][]"
    UINT_GRID = "uint[][]"
    DATE = "date"
    DATETIME = "datetime"
    NONE = "none"
    STATE = "State"


#
# Shortcuts to creating dataclass field metadata
#
rw_float: partial = partial(
    field_config, value_type=ValueType.FLOAT, access_mode=AccessMode.READ_WRITE
)
ro_float: partial = partial(
    field_config, value_type=ValueType.FLOAT, access_mode=AccessMode.READ_ONLY
)
rw_int: partial = partial(
    field_config, value_type=ValueType.INT, access_mode=AccessMode.READ_WRITE
)
ro_int: partial = partial(
    field_config, value_type=ValueType.INT, access_mode=AccessMode.READ_ONLY
)
rw_uint: partial = partial(
    field_config, value_type=ValueType.UINT, access_mode=AccessMode.READ_WRITE
)
ro_uint: partial = partial(
    field_config, value_type=ValueType.UINT, access_mode=AccessMode.READ_ONLY
)
rw_str: partial = partial(
    field_config, value_type=ValueType.STRING, access_mode=AccessMode.READ_WRITE
)
ro_str: partial = partial(
    field_config, value_type=ValueType.STRING, access_mode=AccessMode.READ_ONLY
)
rw_bool: partial = partial(
    field_config, value_type=ValueType.BOOL, access_mode=AccessMode.READ_WRITE
)
ro_bool: partial = partial(
    field_config, value_type=ValueType.BOOL, access_mode=AccessMode.READ_ONLY
)
rw_float_grid: partial = partial(
    field_config,
    value_type=ValueType.FLOAT_GRID,
    access_mode=AccessMode.READ_WRITE,
)
rw_uint_grid: partial = partial(
    field_config,
    value_type=ValueType.UINT_GRID,
    access_mode=AccessMode.READ_WRITE,
)
ro_date: partial = partial(
    field_config, value_type=ValueType.DATE, access_mode=AccessMode.READ_ONLY
)
ro_datetime: partial = partial(
    field_config, value_type=ValueType.DATETIME, access_mode=AccessMode.READ_ONLY
)
ro_state: partial = partial(
    field_config, value_type=ValueType.STATE, access_mode=AccessMode.READ_ONLY
)
rw_state: partial = partial(
    field_config, value_type=ValueType.STATE, access_mode=AccessMode.READ_WRITE
)
ro_str_list: partial = partial(
    field_config, value_type=ValueType.STR_LIST, access_mode=AccessMode.READ_ONLY
)


@order(["access_mode", "allowed_values", "max", "min", "unit", "value", "value_type"])
@with_fields_set
@dataclass
class Value(Generic[T]):
    """Schema for a value to be returned by the API. Most fields are optional."""

    value: T
    value_type: str
    access_mode: str | None = None
    unit: str | None = None
    min: T | None = None
    max: T | None = None
    allowed_values: list[str] | None = None


def construct_value(obj, param):  # noqa: D103
    value = obj[param]["value"]
    meta = obj[param]["metadata"]
    if param == "keys":
        data = serialize(value)
    elif "allowed_values" in meta:
        data = serialize(
            Value(
                value,
                meta["value_type"].value,
                access_mode=meta["access_mode"].value,
                allowed_values=meta["allowed_values"],
            )
        )
    else:
        data = serialize(
            Value(
                value,
                meta["value_type"].value,
                access_mode=meta["access_mode"].value,
            )
        )

    return data


@dataclass
class SequenceComplete:
    """Schema for confirmation returned by operations that do not return values."""

    _sequence_id: int = field(default=1, metadata=skip, init=True, repr=False)

    @serialized("sequence id")  # type: ignore
    @property
    def sequence_id(self) -> int:  # noqa: D102
        return self._sequence_id



import pathlib
from typing import TypedDict

import pydantic.v1.dataclasses
from softioc import builder
from tickit.adapters.epics import EpicsAdapter
from tickit.adapters.io import EpicsIo, TcpIo
from tickit.adapters.specifications import RegexCommand
from tickit.adapters.tcp import CommandAdapter
from tickit.core.adapter import AdapterContainer
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_component import DeviceComponent
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.byte_format import ByteFormat


class SynchrotronTopUpDevice(Device):
    """Device to simulate the top up records from the synchrotron.

    The signal is read via an epics adapter, and set using a tcp adapter.
    """

    class Inputs(TypedDict):
        current: float

    class Outputs(TypedDict):
        countdown: float
        end_countdown: float

    def __init__(
        self,
        initial_countdown: float = 600.0,
        initial_end_countdown: float = 620.0,
        callback_period: int = int(1e9),
        last_current: float = 300.0,
        target_current: float = 300.0,
        minimum_current: float = 270.0,
    ) -> None:
        """Initialise the SynchrotonTopUp device.

        Args:
            initial_countdown (float): Length of time in seconds to deplete
                charge from target_current to minimum_current.
            initial_end_countdown (float): Length of time in seconds to topup the charge
                from minimum_current to target_current.
            callback_period (int): number of nanoseconds to wait before calling again
            last_current (float): Can be set in case the user doesn't want to assume
                the simulation starts at a current of 300mA.
            target_current: The current the device will be topped up to.
            minimum_current (float): The current the synchrotron can fall
                to before topup.
        """
        # Countdown is typically 10 minutes
        self.countdown = self.initial_countdown = initial_countdown

        self.default_fill_time = initial_end_countdown - initial_countdown
        self.target_current = target_current
        self.minimum_current = minimum_current

        # End countdown is typically 10 minutes of countdown, then 10 seconds of fill
        self.end_countdown = self.initial_end_countdown = initial_end_countdown
        self.fill_time = initial_end_countdown - initial_countdown

        self.callback_period = callback_period

        self.last_current = last_current
        self.last_time = None
        self.topup_fill = False

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """Update method that outputs top up record values.

        The device is only altered by adapters so take no inputs.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the value of the machine
                top up records.
        """
        new_current = inputs["current"]

        current_difference = new_current - self.last_current

        topup_fill = current_difference > 0

        # calculate the number of updates to go until topup stage begins or
        # ends: based upon the change of the current.
        updates_to_go = abs(
            topup_fill * (self.target_current - new_current) / current_difference
            + (not topup_fill)
            * (new_current - self.minimum_current)
            / current_difference
        )

        period = self.callback_period * 1e-9
        if self.last_time:
            period = (time - self.last_time) * 1e-9

        self.countdown = (not topup_fill) * (updates_to_go * period)

        # assume topup fill will take 15 seconds, and calculate it directly during
        self.fill_time = topup_fill * (updates_to_go * period) + (
            not topup_fill * self.default_fill_time
        )
        self.end_countdown = self.countdown + self.fill_time

        self.last_current = new_current
        self.last_time = time
        call_at = SimTime(time + self.callback_period)
        return DeviceUpdate(
            SynchrotronTopUpDevice.Outputs(
                countdown=self.countdown,
                end_countdown=self.end_countdown,
            ),
            call_at,
        )

    def get_countdown(self) -> float:
        """Return value of countdown, required for epics adapter."""
        return self.countdown

    def get_end_countdown(self) -> float:
        """Return value of end_countdown, required for epics adapter."""
        return self.end_countdown


class SynchrotronTopUpTCPAdapter(CommandAdapter):
    """A TCP adapter to set a SynchrotronTopUpDevice PV values."""

    device: SynchrotronTopUpDevice
    _byte_format: ByteFormat = ByteFormat(b"%b\r\n")

    def __init__(self, device: SynchrotronTopUpDevice) -> None:
        super().__init__()
        self.device = device

    @RegexCommand(r"CD=(\d+\.?\d*)", interrupt=True, format="utf-8")
    async def set_countdown(self, value: float) -> None:
        """Regex string command to set the value of countdown.

        Args:
            value (int): The new value of countdown.
        """
        self.device.countdown = value

    @RegexCommand(r"ECD=(\d+\.?\d*)", interrupt=True, format="utf-8")
    async def set_end_countdown(self, value: float) -> None:
        """Regex string command to set the value of end_countdown.

        Args:
            value (int): The new value of end_countdown.
        """
        self.device.end_countdown = value

    @RegexCommand(r"CD\?", format="utf-8")
    async def get_countdown(self) -> bytes:
        """Regex string command that returns countdown.

        Returns:
            bytes: The utf-8 encoded value of countdown.
        """
        return str(self.device.countdown).encode("utf-8")

    @RegexCommand(r"ECD\?", format="utf-8")
    async def get_end_countdown(self) -> bytes:
        """Regex string command that returns end_countdown.

        Returns:
            bytes: The utf-8 encoded value of end_countdown.
        """
        return str(self.device.end_countdown).encode("utf-8")


class SynchrotronTopUpEpicsAdapter(EpicsAdapter):
    """Epics adapter for reading device values as a PV through channel access."""

    device: SynchrotronTopUpDevice

    def __init__(self, device: SynchrotronTopUpDevice) -> None:
        super().__init__()
        self.device = device

    def on_db_load(self) -> None:
        """Link epics records with getters for device."""
        self.link_input_on_interrupt(
            builder.aIn("COUNTDOWN"), self.device.get_countdown
        )
        self.link_input_on_interrupt(
            builder.aIn("ENDCOUNTDN"),
            self.device.get_end_countdown,
        )


@pydantic.v1.dataclasses.dataclass
class SynchrotronTopUp(ComponentConfig):
    """Synchrotron top up status component."""

    initial_countdown: float = 600
    initial_end_countdown: float = 620
    callback_period: int = int(1e9)
    host: str = "localhost"
    port: int = 25565
    db_file: str = str(pathlib.Path(__file__).parent.absolute() / "db_files/FILL.db")
    ioc_name: str = "SR-CS-FILL-01"

    def __call__(self) -> Component:  # noqa: D102
        device = SynchrotronTopUpDevice(
            self.initial_countdown,
            self.initial_end_countdown,
            callback_period=self.callback_period,
        )
        adapters = [
            AdapterContainer(
                SynchrotronTopUpTCPAdapter(device),
                TcpIo(
                    self.host,
                    self.port,
                ),
            ),
            AdapterContainer(
                SynchrotronTopUpEpicsAdapter(device),
                EpicsIo(
                    self.ioc_name,
                    self.db_file,
                ),
            ),
        ]
        return DeviceComponent(
            name=self.name,
            device=device,
            adapters=adapters,
        )



from functools import partial, reduce
from operator import and_, or_
from typing import TypedDict

import pydantic.v1.dataclasses
from tickit.core.components.device_component import DeviceComponent

from tickit_devices.zebra._common import Block, BlockConfig, extract_bit


class AndOrBlock(Block):
    """
    Represents an AND or OR gate with 4 inputs.
    If an input is not enabled, it will always be considered False,
    unless it is also inverted.
    """

    class Inputs(TypedDict):
        INP1: bool
        INP2: bool
        INP3: bool
        INP4: bool

    class Outputs(TypedDict):
        OUT: bool

    def __init__(self, name: str):
        super().__init__(name=name, previous_outputs=self.Outputs(OUT=False))

    def _get_input(self, inputs: dict[str, bool], i: int) -> bool:
        if not self.params:
            raise ValueError
        enabled = extract_bit(self.params, f"{self.name}_ENA", i)
        inverted = extract_bit(self.params, f"{self.name}_INV", i)
        return enabled & inputs.get(f"INP{i + 1}", False) ^ inverted

    def _get_next_outputs(self, inputs: Inputs) -> Outputs:
        op = and_ if self.name.startswith("AND") else or_
        get_input = partial(
            self._get_input,
            {
                k: v
                for k, v in inputs.items()
                if isinstance(k, str) and isinstance(v, bool)
            },
        )
        outputs = self.Outputs(OUT=reduce(op, map(get_input, range(4))))
        self.last_input = inputs
        return outputs


@pydantic.v1.dataclasses.dataclass
class AndOrBlockConfig(BlockConfig):
    def __call__(self) -> DeviceComponent:
        return DeviceComponent(name=self.name, device=AndOrBlock(name=self.name))



import itertools
from unittest.mock import ANY, MagicMock, Mock

import pytest
from tickit.core.typedefs import SimTime

from tickit_devices.eiger.eiger import EigerDevice
from tickit_devices.eiger.eiger_status import State
from tickit_devices.eiger.stream.eiger_stream import EigerStream


@pytest.fixture
def mock_stream() -> EigerStream:
    return MagicMock(EigerStream)


@pytest.fixture
def eiger(mock_stream: EigerStream) -> EigerDevice:
    return EigerDevice(stream=mock_stream)


def test_starting_state_is_na(eiger: EigerDevice):
    assert_in_state(eiger, State.NA)


@pytest.mark.asyncio
async def test_initialize(eiger: EigerDevice):
    await eiger.initialize()
    assert_in_state(eiger, State.IDLE)


@pytest.mark.asyncio
async def test_arm(eiger: EigerDevice):
    await eiger.initialize()
    await eiger.arm()
    assert_in_state(eiger, State.READY)


@pytest.mark.asyncio
async def test_disarm(eiger: EigerDevice):
    await eiger.initialize()
    await eiger.arm()
    await eiger.disarm()
    assert_in_state(eiger, State.IDLE)


@pytest.mark.asyncio
async def test_trigger_in_ints_mode_sets_acquire(eiger: EigerDevice):
    await eiger.initialize()
    eiger.settings.trigger_mode = "ints"
    await eiger.arm()
    assert_in_state(eiger, State.READY)
    await eiger.trigger()
    assert_in_state(eiger, State.ACQUIRE)


@pytest.mark.asyncio
async def test_trigger_in_ints_mode_while_not_armed_is_ignored(
    eiger: EigerDevice,
):
    await eiger.initialize()
    eiger.settings.trigger_mode = "ints"
    await eiger.trigger()
    assert_in_state(eiger, State.IDLE)


@pytest.mark.asyncio
async def test_trigger_in_exts_mode_is_ignored(eiger: EigerDevice):
    await eiger.initialize()
    eiger.settings.trigger_mode = "exts"
    await eiger.arm()
    assert_in_state(eiger, State.READY)
    await eiger.trigger()
    eiger.update(SimTime(0.0), {})
    assert_in_state(eiger, State.READY)


@pytest.mark.asyncio
async def test_update_in_exts_mode_is_ignored(eiger: EigerDevice):
    await eiger.initialize()
    eiger.settings.trigger_mode = "exts"
    await eiger.arm()
    assert_in_state(eiger, State.READY)
    eiger.update(SimTime(0.0), {})
    assert_in_state(eiger, State.READY)


@pytest.mark.asyncio
async def test_trigger_in_exts_mode_while_not_armed_is_ignored(
    eiger: EigerDevice,
):
    await eiger.initialize()
    eiger.settings.trigger_mode = "exts"
    await eiger.trigger()
    assert_in_state(eiger, State.IDLE)


@pytest.mark.asyncio
async def test_cancel(eiger: EigerDevice):
    await eiger.cancel()
    assert_in_state(eiger, State.READY)


@pytest.mark.asyncio
async def test_abort(eiger: EigerDevice):
    await eiger.abort()
    assert_in_state(eiger, State.IDLE)


@pytest.mark.asyncio
async def test_armed_eiger_starts_series(eiger: EigerDevice, mock_stream: Mock):
    await eiger.initialize()
    eiger.settings.trigger_mode = "ints"
    await eiger.arm()
    mock_stream.begin_series.assert_called_once_with(eiger.settings, 1, "basic")


@pytest.mark.asyncio
async def test_disarmed_eiger_starts_and_ends_series(
    eiger: EigerDevice, mock_stream: Mock
):
    await eiger.initialize()
    eiger.settings.trigger_mode = "ints"
    await eiger.arm()
    await eiger.disarm()
    mock_stream.begin_series.assert_called_once_with(eiger.settings, 1, "basic")
    mock_stream.end_series.assert_called_once_with(1)


@pytest.mark.asyncio
async def test_cancelled_eiger_starts_and_ends_series(
    eiger: EigerDevice, mock_stream: Mock
):
    await eiger.initialize()
    eiger.settings.trigger_mode = "ints"
    await eiger.arm()
    await eiger.cancel()
    mock_stream.begin_series.assert_called_once_with(eiger.settings, 1, "basic")
    mock_stream.end_series.assert_called_once_with(1)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "num_frames,num_series", list(itertools.product([0, 1, 2, 10], [1, 2, 3]))
)
async def test_acquire_frames_in_ints_mode(
    eiger: EigerDevice,
    mock_stream: Mock,
    num_frames: int,
    num_series: int,
):
    for series in range(1, num_series + 1):
        await eiger.initialize()
        eiger.settings.trigger_mode = "ints"
        eiger.settings.nimages = num_frames
        await eiger.arm()
        await eiger.trigger()

        # Extra update cleans up state
        for i in range(num_frames):
            update = eiger.update(SimTime(i), {})
            assert update.call_at == SimTime(i + int(0.12 * 1e9))

        update = eiger.update(SimTime(0.0), {})
        assert update.call_at is None

        mock_stream.begin_series.assert_called_with(eiger.settings, series, "basic")
        assert mock_stream.begin_series.call_count == series
        if num_frames > 0:
            mock_stream.insert_image.assert_called_with(ANY, series)
        assert mock_stream.insert_image.call_count == series * num_frames
        mock_stream.end_series.assert_called_with(series)
        assert mock_stream.end_series.call_count == series

        assert_in_state(eiger, State.IDLE)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "num_frames,num_series", list(itertools.product([0, 1, 2, 10], [1, 2, 3]))
)
async def test_acquire_frames_in_exts_mode(
    eiger: EigerDevice,
    mock_stream: Mock,
    num_frames: int,
    num_series: int,
):
    for series in range(1, num_series + 1):
        await eiger.initialize()
        eiger.settings.trigger_mode = "exts"
        eiger.settings.nimages = num_frames
        await eiger.arm()

        # Trigger detector
        update = eiger.update(SimTime(0.0), {"trigger": True})
        assert update.call_at == 0.0

        # Extra update cleans up state
        for i in range(num_frames):
            update = eiger.update(SimTime(i), {})
            assert update.call_at == SimTime(i + int(0.12 * 1e9))

        update = eiger.update(SimTime(0.0), {})
        assert update.call_at is None

        mock_stream.begin_series.assert_called_with(eiger.settings, series, "basic")
        assert mock_stream.begin_series.call_count == series
        if num_frames > 0:
            mock_stream.insert_image.assert_called_with(ANY, series)
        assert mock_stream.insert_image.call_count == series * num_frames
        mock_stream.end_series.assert_called_with(series)
        assert mock_stream.end_series.call_count == series

        assert_in_state(eiger, State.IDLE)


@pytest.mark.asyncio
@pytest.mark.parametrize("num_series", [1, 2, 3])
async def test_abort_mid_acquisition(
    eiger: EigerDevice,
    mock_stream: Mock,
    num_series: int,
):
    for series in range(1, num_series + 1):
        await eiger.initialize()
        eiger.settings.trigger_mode = "ints"
        eiger.settings.nimages = 3
        await eiger.arm()
        await eiger.trigger()

        eiger.update(SimTime(0.0), {})
        eiger.update(SimTime(0.0), {})

        await eiger.abort()

        eiger.update(SimTime(0.0), {})

        mock_stream.end_series.assert_called_with(series)
        assert_in_state(eiger, State.IDLE)


def assert_in_state(eiger: EigerDevice, state: State) -> None:
    assert state is eiger.get_state()



import asyncio
import logging
import struct

import numpy as np
import pytest
from tickit.core.device import DeviceUpdate
from tickit.core.typedefs import SimTime

from tickit_devices.cryostream.cryostream import CryostreamDevice
from tickit_devices.cryostream.states import PhaseIds
from tickit_devices.cryostream.status import ExtendedStatus

# # # # # Cryostream Tests # # # # #


@pytest.fixture
def cryostream() -> CryostreamDevice:
    return CryostreamDevice()


def test_cryostream_constructor():
    CryostreamDevice()


def test_cryostream_update_hold(cryostream):
    starting_temperature = cryostream.gas_temp
    device_update: DeviceUpdate = cryostream.update(time=SimTime(1e9), inputs={})
    assert device_update.outputs["temperature"] == starting_temperature


@pytest.mark.asyncio
async def test_cryostream_update_cool(cryostream: CryostreamDevice):
    starting_temperature = cryostream.gas_temp
    target_temperature = starting_temperature - 50
    await cryostream.cool(target_temperature)

    time = SimTime(0)
    device_update: DeviceUpdate
    while cryostream.phase_id != PhaseIds.HOLD.value:
        logging.info(f"Running time step: {time}")
        device_update = cryostream.update(time, inputs={})
        time_update: SimTime | None = device_update.call_at

        if time_update is None:
            time = SimTime(int(time) + int(1e9))
            continue
        else:
            time = time_update

    max_diff = 10
    margin_of_error = np.array([+max_diff, -max_diff])
    assert any(
        (device_update.outputs["temperature"] + margin_of_error) == target_temperature
    )


@pytest.mark.asyncio
async def test_cryostream_update_end(cryostream: CryostreamDevice):
    starting_temperature = cryostream.default_temp_shutdown - 100
    cryostream.gas_temp = starting_temperature
    await cryostream.end(cryostream.default_ramp_rate)
    assert cryostream.phase_id == PhaseIds.RAMP.value
    assert cryostream.gas_flow == 5

    time = SimTime(0)
    device_update: DeviceUpdate
    while cryostream.phase_id != PhaseIds.HOLD.value:
        logging.info(f"Running time step: {time}")
        device_update = cryostream.update(time, inputs={})
        time_update: SimTime | None = device_update.call_at

        if time_update is None:
            continue
        else:
            time = time_update

    max_diff = 10
    target = cryostream.default_temp_shutdown
    assert (
        target - max_diff <= device_update.outputs["temperature"] <= target + max_diff
    )


@pytest.mark.asyncio
async def test_cryostream_update_plat(cryostream: CryostreamDevice):
    starting_temperature = cryostream.gas_temp
    await cryostream.plat(5)
    assert cryostream.phase_id == PhaseIds.PLAT.value

    time = SimTime(0)
    time_update: SimTime | None = time
    device_update: DeviceUpdate
    while time_update is not None:
        logging.info(f"Running time step: {time}")
        device_update = cryostream.update(time, inputs={})
        time_update = device_update.call_at

        if time_update is None:
            continue
        else:
            time = time_update

    assert device_update.outputs["temperature"] == starting_temperature


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tickit_task", ["examples/configs/cryostream/cryo-tcp.yaml"], indirect=True
)
async def test_cryostream_system(tickit_task):
    reader, writer = await asyncio.open_connection("localhost", 25565)

    async def write(data: bytes):
        writer.write(data)
        await writer.drain()

    async def get_status() -> ExtendedStatus:
        return ExtendedStatus.from_packed(await reader.read(42))

    # Check we start at 300K
    status = await get_status()
    assert status.gas_temp == 30000
    assert status.turbo_mode == 0
    # Start a ramp
    await write(b"\x06\x0b" + struct.pack(">HH", 360, 30030))
    # Check we go to to 300.3K, eventually
    status = await get_status()
    assert status.gas_temp == pytest.approx(30015, rel=5)
    status = await get_status()
    assert status.gas_temp == pytest.approx(30025, rel=5)
    status = await get_status()
    assert status.gas_temp == pytest.approx(30030, rel=5)
    # Turbo on
    await write(b"\x03\x14\x01")
    status = await get_status()
    assert status.gas_temp == pytest.approx(30030, rel=5)
    assert status.turbo_mode == 1
    # Cool to 300K
    await write(b"\x04\x0e" + struct.pack(">H", 30000))
    status = await get_status()
    assert status.gas_temp == pytest.approx(30000, rel=5)

    writer.close()
    await writer.wait_closed()



import asyncio

from tickit.adapters.specifications import RegexCommand
from tickit.adapters.system import BaseSystemSimulationAdapter
from tickit.core.components.device_component import DeviceComponent
from tickit.core.management.event_router import InverseWiring, Wiring
from tickit.core.typedefs import ComponentID

from tickit_devices.zebra._common import param_types, register_names


class ZebraAdapter(BaseSystemSimulationAdapter):
    _components: dict[ComponentID, DeviceComponent]
    params: dict[str, int]
    """
    Network adapter for a Zebra system simulation, which operates a TCP server for
    reading and setting configuration of blocks and internal wiring mapping.
    N.B. Reading and setting internal wiring is not currently supported,and all
    "mux" related queries will return as though successful but not operate.

    See documentation for the Zebra:
    `https://github.com/dls-controls/zebra/blob/master/documentation/TDI-CTRL-TNO-042-Zebra-Manual.pdf`

    Configuration that is currently supported:
    - For AND/OR gates N=1,2,3,4, the following (default 0) may be set
    to Σ2**M where M is each input (1,2,3,4) for which the behaviour is desired.
    {AND|OR}{N}_INV: Inverts input(s) M
    {AND|OR}{N}_ENA: Enables input(s) M
    """

    def __init__(self, params: dict[str, int]):
        self.params = params

    def setup_adapter(
        self,
        components: dict[ComponentID, DeviceComponent],
        wiring: Wiring | InverseWiring,
    ) -> None:
        """
        Sets the shared configuration/"params" between the Zebra and its components
        then instantiates them.
        """
        for block in components.values():
            block.device.params = self.params
        super().setup_adapter(components, wiring)

    @RegexCommand(rb"W([0-9A-F]{2})([0-9A-F]{4})\n", interrupt=True)
    async def set_reg(self, reg: str, value: str) -> bytes:
        reg_int, value_int = int(reg, base=16), int(value, base=16)
        reg_name = register_names[reg_int]

        if reg_name in self.params:
            self.params[reg_name] = value_int

            for block_name in param_types[reg_name].blocks:
                await asyncio.create_task(
                    self._components[block_name].raise_interrupt()
                )

        else:
            self._set_mux(reg_name, value_int)

        return b"W%02XOK" % reg_int

    @RegexCommand(rb"R([0-9A-F]{2})\n")
    async def get_reg(self, reg: str) -> bytes:
        reg_int = int(reg, base=16)
        reg_name = register_names[reg_int]
        try:
            value_int = self.params[reg_name]
        except KeyError:
            value_int = self._read_mux(reg_name)
        return b"R%02X%04XOK" % (reg_int, value_int)

    def _read_mux(self, reg_name: str) -> int:  # type: ignore
        # TODO: Support reading Muxes: map internal wiring into Mux values
        return 0

    def _set_mux(self, reg_name: str, value: int) -> int:  # type: ignore
        # TODO: Support setting Muxes: map Mux value into internal wiring
        # TODO: Support cyclic wirings involving Zebra: AND1_OUT->AND1_INP1 is valid
        return 0



from dataclasses import dataclass, field, fields
from typing import Any

from tickit_devices.eiger.eiger_schema import ro_int, ro_str, ro_uint


def monitor_status_keys() -> list[str]:
    return ["buffer_free", "dropped", "error", "state"]


@dataclass
class MonitorStatus:
    """Eiger monitor status taken from the API spec."""

    error: list[str] = field(default_factory=lambda: [], metadata=ro_str())
    buffer_fill_level: int = field(default=2, metadata=ro_int())
    buffer_free: int = field(default_factory=int, metadata=ro_uint())
    dropped: int = field(default=0, metadata=ro_uint())
    state: str = field(
        default="normal", metadata=ro_str(allowed_values=["normal", "overflow"])
    )
    keys: list[str] = field(default_factory=monitor_status_keys)

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        for field_ in fields(self):
            if field_.name == key:
                return {"value": vars(self)[field_.name], "metadata": field_.metadata}
        raise ValueError(f"No field with name {key}")



from contextlib import nullcontext as does_not_raise
from random import getrandbits
from typing import Any
from unittest.mock import Mock

import pytest

from tickit_devices.cryostream.base import CryostreamBase
from tickit_devices.cryostream.states import AlarmCodes, PhaseIds, RunModes
from tickit_devices.cryostream.status import ExtendedStatus, Status


def rand_bool() -> bool:
    return bool(getrandbits(1))


@pytest.mark.asyncio
async def test_restart():
    cryostream_base = CryostreamBase()
    await cryostream_base.restart()
    assert cryostream_base.run_mode in (
        RunModes.STARTUPOK.value,
        RunModes.STARTUPFAIL.value,
    )


@pytest.mark.parametrize(
    "test_params",
    [
        pytest.param(
            {
                "starting_temperature": 10000,
                "target_temperature": 9000,
                "ramp_rate": 360,
                "expected_phase_id": PhaseIds.RAMP.value,
                "expected_alarm_code": AlarmCodes.NO_ERRORS.value,
                "expected_gas_flow": 5,
            },
            id="normal params",
        ),
        pytest.param(
            {
                "starting_temperature": 8999,
                "target_temperature": 8000,
                "ramp_rate": 360,
                "expected_phase_id": PhaseIds.RAMP.value,
                "expected_alarm_code": AlarmCodes.NO_ERRORS.value,
                "expected_gas_flow": 10,
            },
            id="gas gas gas",
        ),
        pytest.param(
            {
                "starting_temperature": 10000,
                "target_temperature": 90001,
                "ramp_rate": 360,
                "expected_phase_id": PhaseIds.RAMP.value,
                "expected_alarm_code": AlarmCodes.TEMP_CONTROL_ERROR.value,
                "expected_gas_flow": 10,
            },
            id="target temperature too high",
        ),
        pytest.param(
            {
                "starting_temperature": 10000,
                "target_temperature": 9000,
                "ramp_rate": 361,
                "expected_phase_id": PhaseIds.RAMP.value,
                "expected_alarm_code": AlarmCodes.TEMP_CONTROL_ERROR.value,
                "expected_gas_flow": 5,
            },
            id="ramp rate too high",
        ),
    ],
)
@pytest.mark.asyncio
async def test_ramp(test_params):
    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = test_params["starting_temperature"]

    await cryostream_base.ramp(
        ramp_rate=test_params["ramp_rate"],
        target_temp=test_params["target_temperature"],
    )

    assert cryostream_base.run_mode == RunModes.RUN.value
    assert cryostream_base.alarm_code == test_params["expected_alarm_code"]
    assert cryostream_base.phase_id == test_params["expected_phase_id"]
    assert cryostream_base.gas_flow == test_params["expected_gas_flow"]


@pytest.mark.parametrize(
    "test_params",
    [
        pytest.param(
            {
                "duration": 1440,
                "expected_phase_id": PhaseIds.PLAT.value,
                "expected_run_mode": RunModes.RUN.value,
                "raises": does_not_raise(),
            },
            id="normal params",
        ),
        pytest.param(
            {
                "duration": 0,
                "expected_phase_id": None,
                "expected_run_mode": None,
                "raises": pytest.raises(ValueError),
            },
            id="zero duration",
        ),
        pytest.param(
            {
                "duration": 1441,
                "expected_phase_id": None,
                "expected_run_mode": None,
                "raises": pytest.raises(ValueError),
            },
            id="too long",
        ),
    ],
)
@pytest.mark.asyncio
async def test_plat(test_params):
    cryostream_base = CryostreamBase()
    initial_run_mode = cryostream_base.run_mode
    initial_phase_id = cryostream_base.phase_id
    with test_params["raises"]:
        await cryostream_base.plat(test_params["duration"])

    if test_params["expected_phase_id"] is None:
        assert cryostream_base.phase_id == initial_phase_id
    else:
        assert cryostream_base.phase_id == test_params["expected_phase_id"]

    if test_params["expected_run_mode"] is None:
        assert cryostream_base.run_mode == initial_run_mode
    else:
        assert cryostream_base.run_mode == test_params["expected_run_mode"]


@pytest.mark.asyncio
async def test_hold():
    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = 28000
    await cryostream_base.hold()
    final_temperature = cryostream_base.update_temperature(100e9)
    assert final_temperature == 28000
    assert cryostream_base.run_mode == RunModes.RUN.value
    assert cryostream_base.phase_id == PhaseIds.HOLD.value


@pytest.mark.asyncio
async def test_cool():
    cryostream_base = CryostreamBase()
    cryostream_base.ramp = Mock(cryostream_base.ramp)
    starting_temperature = cryostream_base.gas_temp
    target_temperature = starting_temperature - 10 * 5
    await cryostream_base.cool(target_temperature)
    assert cryostream_base.run_mode == RunModes.RUN.value
    assert cryostream_base.phase_id == PhaseIds.COOL.value
    cryostream_base.ramp.assert_awaited_once_with(
        cryostream_base.default_ramp_rate, target_temperature
    )


@pytest.mark.parametrize(
    "test_params",
    [
        pytest.param(
            {
                "initial_gas_temp": CryostreamBase.default_temp_shutdown,
                "expected_run_mode": RunModes.SHUTDOWNOK.value,
                "expected_gas_flow": {0},
            },
            id="shutdown okay",
        ),
        pytest.param(
            {
                "initial_gas_temp": CryostreamBase.default_temp_shutdown - 100,
                "expected_run_mode": RunModes.SHUTDOWNFAIL.value,
                "expected_gas_flow": {5, 10},
            },
            id="shutdown fails",
        ),
    ],
)
@pytest.mark.asyncio
async def test_end(test_params: dict[str, Any]):
    def set_gas_flow(value: int, target_temp: int) -> None:
        if rand_bool():
            cryostream_base.gas_flow = 10
        else:
            cryostream_base.gas_flow = 5

    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = test_params["initial_gas_temp"]
    cryostream_base.ramp = Mock(  # type: ignore
        cryostream_base.ramp, side_effect=set_gas_flow
    )
    await cryostream_base.end(ramp_rate=cryostream_base.max_rate)
    assert cryostream_base.phase_id == PhaseIds.END.value
    assert cryostream_base.run_mode == test_params["expected_run_mode"]
    assert cryostream_base.gas_flow in test_params["expected_gas_flow"]


@pytest.mark.parametrize(
    "test_params",
    [
        pytest.param(
            {
                "initial_gas_temp": CryostreamBase.default_temp_shutdown,
                "expected_run_mode": RunModes.SHUTDOWNOK.value,
            },
            id="shutdown okay",
        ),
        pytest.param(
            {
                "initial_gas_temp": CryostreamBase.default_temp_shutdown - 100,
                "expected_run_mode": RunModes.SHUTDOWNFAIL.value,
            },
            id="shutdown fails",
        ),
    ],
)
@pytest.mark.asyncio
async def test_purge(test_params: dict[str, Any]):
    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = test_params["initial_gas_temp"]
    cryostream_base.ramp = Mock(cryostream_base.ramp)  # type: ignore
    await cryostream_base.purge()
    assert cryostream_base.phase_id == PhaseIds.PURGE.value
    assert cryostream_base.run_mode == test_params["expected_run_mode"]


@pytest.mark.asyncio
async def test_stop():
    cryostream_base = CryostreamBase()
    await cryostream_base.stop()
    assert cryostream_base.gas_flow == 0
    assert cryostream_base._target_temp == cryostream_base.gas_temp
    assert cryostream_base.run_mode == RunModes.SHUTDOWNOK.value


@pytest.mark.parametrize(
    "test_params",
    argvalues=[
        {"temp": 350, "expected_gas_flow": 5},
        {"temp": 300, "expected_gas_flow": 10},
    ],
)
@pytest.mark.asyncio
async def test_turbo(test_params):
    cryostream_base = CryostreamBase()
    cryostream_base.gas_temp = test_params["temp"]
    await cryostream_base.turbo(1)
    assert cryostream_base.turbo_mode == 1
    assert cryostream_base.gas_flow == test_params["expected_gas_flow"]


def test_update_temperature():
    cryostream_base = CryostreamBase()
    cryostream_base._target_temp = cryostream_base.gas_temp - cryostream_base.gas_flow

    cryostream_base.update_temperature(1e9)
    assert cryostream_base.gas_temp == cryostream_base._target_temp


@pytest.mark.asyncio
async def test_set_status_format():
    cryostream_base = CryostreamBase()
    await cryostream_base.set_status_format(0)
    assert isinstance(cryostream_base.status, Status)

    await cryostream_base.set_status_format(1)
    assert isinstance(cryostream_base.extended_status, ExtendedStatus)


@pytest.mark.asyncio
async def test_get_status():
    cryostream_base = CryostreamBase()
    status = await cryostream_base.get_status(0)
    assert isinstance(status, Status)

    extended_status = await cryostream_base.get_status(1)
    assert isinstance(extended_status, ExtendedStatus)



import pytest

from tickit_devices.eiger.stream.stream_config import StreamConfig

# # # # # Eiger StreamConfig Tests # # # # #


@pytest.fixture
def stream_config() -> StreamConfig:
    return StreamConfig()


def test_eiger_stream_config_constructor():
    StreamConfig()


def test_eiger_stream_config_getitem(stream_config):
    assert "enabled" == stream_config["mode"]["value"]
    stream_config["mode"] = "disabled"
    assert "disabled" == stream_config["mode"]["value"]

    with pytest.raises(ValueError):
        stream_config["doesnt_exist"]



import pytest
from tickit.core.device import DeviceUpdate
from tickit.core.typedefs import SimTime

from tickit_devices.synchrotron.synchrotron_topup import SynchrotronTopUpDevice

COUNTDOWN = 600
ENDCOUNTDN = 620


@pytest.fixture
def synchrotron_topup() -> SynchrotronTopUpDevice:
    return SynchrotronTopUpDevice(
        initial_countdown=COUNTDOWN,
        initial_end_countdown=ENDCOUNTDN,
        callback_period=int(1e9),
    )


def test_synchrotron_topup_constructor(
    synchrotron_topup: SynchrotronTopUpDevice,
):
    pass


def test_synchrotron_topup_get_countdown(
    synchrotron_topup: SynchrotronTopUpDevice,
):
    synchrotron_topup.countdown = COUNTDOWN
    assert synchrotron_topup.get_countdown() == COUNTDOWN


def test_synchrotron_topup_update_new_countdown(
    synchrotron_topup: SynchrotronTopUpDevice,
):
    device_update: DeviceUpdate = synchrotron_topup.update(
        SimTime(0), inputs={"current": 299.95}
    )

    countdown = (299.95 - 270) / 0.05

    # countdown seconds need to be rounded since the scheduler introduces slight changes
    # many decimal places in
    assert round(device_update.outputs["countdown"], -7) == round(countdown, -7)
    assert round(device_update.outputs["end_countdown"], -7) == round(
        countdown + 15, -7
    )


def test_synchrotron_topup_update_new_countdown_topup_fill():
    synchrotron_topup = SynchrotronTopUpDevice(last_current=290)
    device_update: DeviceUpdate = synchrotron_topup.update(
        SimTime(0), inputs={"current": 291}
    )

    countdown = (300 - 291) / (291 - 290)

    # countdown seconds need to be rounded since the scheduler introduces slight changes
    # many decimal places in
    assert device_update.outputs["countdown"] == 0
    assert device_update.outputs["end_countdown"] == countdown


def test_synchrotron_topup_call_at(synchrotron_topup: SynchrotronTopUpDevice):
    device_update: DeviceUpdate = synchrotron_topup.update(
        SimTime(0), inputs={"current": 289}
    )
    expected_output = int(1e9)
    assert device_update.call_at == expected_output



from typing import TypedDict

from softioc import builder
from tickit.adapters.epics import EpicsAdapter
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime


class PneumaticDevice(Device):
    """Pneumatic Device with movement controls."""

    class Inputs(TypedDict): ...

    #: A typed mapping containing the current output value
    class Outputs(TypedDict):
        output: float

    def __init__(
        self,
        initial_speed: float,
        initial_state: bool,
    ) -> None:
        """Initialise a Pneumatic object."""
        self.speed: float = initial_speed
        self.state: bool = initial_state
        self.moving: bool = False
        self.time_at_last_update: float = 0.0

    def set_speed(self, speed: float) -> None:
        """Set the speed of movement for the device.

        Args:
            speed (float): The desired movement speed of the device.
        """
        self.speed = speed

    def get_speed(self) -> float:
        """Get the speed of movement of the device."""
        return self.speed

    def set_state(self) -> None:
        """Toggles the target state of the device."""
        self.moving = True
        self.target_state = not self.state

    def get_state(self) -> bool:
        """Gets the current state of the device."""
        return self.state

    def update(self, time: SimTime, inputs) -> DeviceUpdate[Outputs]:
        """Run the update logic for the device.

        If the device is moving then the state of the device is updated.
        Otherwise nothing changes. In either case the current state of the
        device is returned.

        Args:
            time (SimTime): The time of the simulation in nanoseconds.
            inputs (dict): The dict containing the input values of state of the device.

        Returns:
            DeviceUpdate: A container for the Device's outputs and a callback time.
        """
        if self.moving:
            callback_period = SimTime(int(1e9 / self.speed))
            self.state = self.target_state
            self.moving = False
            return DeviceUpdate(
                self.Outputs(output=self.state),
                callback_period,
            )
        else:
            return DeviceUpdate(self.Outputs(output=self.state), None)


class PneumaticAdapter(EpicsAdapter):
    """An adapter for the Pneumatic class.

    Connects the device to an external messaging protocol.
    """

    device: PneumaticDevice

    def __init__(self, device: PneumaticDevice) -> None:
        super().__init__()
        self.device = device

    async def callback(self, value) -> None:
        """Set the state of the device and await a response.

        Args:
            value (bool): The value to set the state to.
        """
        self.device.set_state()
        await self.interrupt()

    def on_db_load(self):
        """Adds a record of the current state to the mapping of interrupting records."""
        builder.boolOut("FILTER", initial_value=False, on_update=self.callback)
        self.link_input_on_interrupt(
            builder.boolIn("FILTER_RBV"), self.device.get_state
        )



import pytest

from tickit_devices.eiger.monitor.monitor_config import MonitorConfig

# # # # # Eiger MonitorConfig Tests # # # # #


@pytest.fixture
def monitor_config() -> MonitorConfig:
    return MonitorConfig()


def test_eiger_monitor_config_constructor():
    MonitorConfig()


def test_eiger_monitor_config_get_set(monitor_config):
    assert "enabled" == monitor_config["mode"]["value"]
    monitor_config["mode"] = "disabled"
    assert "disabled" == monitor_config["mode"]["value"]

    with pytest.raises(ValueError):
        monitor_config["doesnt_exist"]



import pytest
from pytest_mock import MockerFixture

from tickit_devices.eiger.eiger import EigerDevice, get_changed_parameters
from tickit_devices.eiger.eiger_adapters import EigerRESTAdapter, EigerZMQAdapter


def test_after_update(mocker: MockerFixture) -> None:
    test_data = [b"data", b"some more data"]

    # Mock consume_data to return with data the first time and nothing the second time
    device_mock = mocker.MagicMock()
    device_mock.stream.consume_data.side_effect = [test_data, []]

    zmq_adapter = EigerZMQAdapter(device_mock.stream)
    add_mock = mocker.patch.object(zmq_adapter, "add_message_to_stream")

    # Test after_update only calls add_message_to_stream with non-empty data
    zmq_adapter.after_update()
    add_mock.assert_called_once_with(test_data)
    add_mock.reset_mock()
    zmq_adapter.after_update()
    add_mock.assert_not_called()


@pytest.mark.asyncio
async def test_rest_adapter_404(mocker: MockerFixture):
    eiger_adapter = EigerRESTAdapter(EigerDevice())

    request = mocker.MagicMock()
    request.json = mocker.AsyncMock()

    request.match_info = {"parameter_name": "doesnt_exist"}
    assert (await eiger_adapter.get_config(request)).status == 404
    assert (await eiger_adapter.get_board_000_status(request)).status == 404

    request.json.return_value = {}
    assert (await eiger_adapter.put_config(request)).status == 404

    request.match_info = {"parameter_name": "doesnt_exist", "threshold": "1"}
    assert (await eiger_adapter.get_threshold_config(request)).status == 404
    assert (await eiger_adapter.put_threshold_config(request)).status == 404

    request.match_info = {"param": "doesnt_exist"}
    assert (await eiger_adapter.get_monitor_status(request)).status == 404
    assert (await eiger_adapter.get_monitor_config(request)).status == 404
    assert (await eiger_adapter.put_monitor_config(request)).status == 404
    assert (await eiger_adapter.get_filewriter_status(request)).status == 404
    assert (await eiger_adapter.get_filewriter_config(request)).status == 404
    assert (await eiger_adapter.put_filewriter_config(request)).status == 404
    assert (await eiger_adapter.get_stream_status(request)).status == 404
    assert (await eiger_adapter.get_stream_config(request)).status == 404
    assert (await eiger_adapter.put_stream_config(request)).status == 404

    request.match_info = {"status_param": "doesnt_exist"}
    assert (await eiger_adapter.get_status(request)).status == 404
    assert (await eiger_adapter.get_builder_status(request)).status == 404


@pytest.mark.asyncio
async def test_rest_adapter_command_404(mocker: MockerFixture):
    eiger_adapter = EigerRESTAdapter(EigerDevice())

    request = mocker.MagicMock()
    request.text = mocker.AsyncMock()
    request.json = mocker.AsyncMock()
    request.match_info = {"key": "value"}

    assert (await eiger_adapter.initialize_eiger(request)).status == 404
    assert (await eiger_adapter.arm_eiger(request)).status == 404
    assert (await eiger_adapter.disarm_eiger(request)).status == 404
    assert (await eiger_adapter.trigger_eiger(request)).status == 404
    assert (await eiger_adapter.cancel_eiger(request)).status == 404
    assert (await eiger_adapter.abort_eiger(request)).status == 404


@pytest.mark.asyncio
async def test_detector_put_responses(mocker: MockerFixture):
    # test special keys have non-trivial response
    custom_response_keys = [
        "auto_summation",
        "count_time",
        "frame_time",
        "flatfield",
        "incident_energy",
        "photon_energy",
        "pixel_mask",
        "threshold/1/flatfield",
        "roi_mode",
        "threshold_energy",
        "threshold/1/energy",
        "threshold/2/energy",
        "threshold/1/mode",
        "threshold/2/mode",
        "threshold/1/pixel_mask",
        "threshold/difference/mode",
    ]

    for key in custom_response_keys:
        assert get_changed_parameters(key) != [key]

    assert get_changed_parameters("phi_start") == ["phi_start"]
    assert get_changed_parameters("threshold/1/pixel_mask") == [
        "pixel_mask",
        "threshold/1/pixel_mask",
    ]
    assert get_changed_parameters("threshold/difference/mode") == ["difference_mode"]

    eiger_adapter = EigerRESTAdapter(EigerDevice())

    request = mocker.MagicMock()
    request.json = mocker.AsyncMock()

    request.match_info = {"parameter_name": "count_time", "value": 1.0}
    response = await eiger_adapter.put_config(request)
    assert response.body == (
        b'["bit_depth_image", "bit_depth_readout", "count_time",'
        b' "countrate_correction_count_cutoff",'
        b' "frame_count_time", "frame_time"]'
    )
    # trivial case just returns the single parameter
    request.match_info = {"parameter_name": "phi_start", "value": 1.0}
    response = await eiger_adapter.put_config(request)
    assert response.body == b'["phi_start"]'

    # test threshold responses work

    request.match_info = {"parameter_name": "mode", "threshold": "1", "value": 1}
    request.json = mocker.AsyncMock(return_value={"value": 1})
    response = await eiger_adapter.put_threshold_config(request)
    assert response.body == b'["threshold/1/mode", "threshold/difference/mode"]'



from dataclasses import dataclass, field, fields
from typing import Any

from tickit_devices.eiger.eiger_schema import ro_str, ro_str_list


@dataclass
class FileWriterStatus:
    """Eiger filewriter status taken from the API spec."""

    state: str = field(default="ready", metadata=ro_str())
    error: list[str] = field(default_factory=lambda: [], metadata=ro_str_list())
    files: list[str] = field(default_factory=lambda: [], metadata=ro_str_list())

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        for field_ in fields(self):
            if field_.name == key:
                return {"value": vars(self)[field_.name], "metadata": field_.metadata}
        raise ValueError(f"No field with name {key}")



import asyncio

import pytest
from aioca import caget, caput
from tickit.core.device import DeviceUpdate
from tickit.core.typedefs import SimTime

from tickit_devices.femto.current import CurrentDevice
from tickit_devices.femto.femto import FemtoDevice


@pytest.fixture
def femto() -> FemtoDevice:
    return FemtoDevice(initial_gain=2.5, initial_current=0.0)


def test_femto_constructor(femto: FemtoDevice):
    pass


def test_set_and_get_gain(femto: FemtoDevice):
    femto.set_gain(3.0)
    assert femto.get_gain() == 3.0


def test_set_and_get_current(femto: FemtoDevice):
    femto.set_current(3.0)
    assert femto.get_current() == 3.0 * femto.get_gain()


def test_femto_update(femto: FemtoDevice):
    device_input = FemtoDevice.Inputs({"input": 3.0})
    time = SimTime(0)
    update: DeviceUpdate = femto.update(time, device_input)
    assert update.outputs["current"] == 3.0 * femto.get_gain()


# # # # # # # # # # CurrentDevice # # # # # # # # # #


@pytest.fixture
def current_device() -> CurrentDevice:
    return CurrentDevice(callback_period=1000000000)


def test_current_device_constructor(current_device: CurrentDevice):
    pass


def test_current_device_update(current_device: CurrentDevice):
    time = SimTime(0)
    device_input = {"bleep": "bloop"}
    update: DeviceUpdate = current_device.update(time, device_input)
    assert 100 <= update.outputs["output"] < 200


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tickit_process", ["examples/configs/femto/current-monitor.yaml"], indirect=True
)
async def test_femto_system(tickit_process):
    assert (await caget("FEMTO:GAIN_RBV")) == 2.5
    current = await caget("FEMTO:CURRENT")
    assert 100 * 2.5 <= current < 200 * 2.5
    await caput("FEMTO:GAIN", 0.01)
    await asyncio.sleep(0.5)
    assert (await caget("FEMTO:GAIN_RBV")) == 0.01
    current = await caget("FEMTO:CURRENT")
    assert 100 * 0.01 <= current < 200 * 0.01



import cbor2
import numpy as np


def decode_multi_dim_array(tag, column_major):
    dimensions, contents = tag.value
    if isinstance(contents, list):
        array = np.empty((len(contents),), dtype=object)
        array[:] = contents
    elif isinstance(contents, np.ndarray | np.generic):
        array = contents
    else:
        raise cbor2.CBORDecodeValueError("expected array or typed array")
    return array.reshape(dimensions, order="F" if column_major else "C")


def decode_typed_array(tag, dtype):
    if not isinstance(tag.value, bytes):
        raise cbor2.CBORDecodeValueError("expected byte string in typed array")
    return np.frombuffer(tag.value, dtype=dtype)


def decode_dectris_compression(tag):
    _algorithm, _elem_size, encoded = tag.value
    return encoded


tag_decoders = {
    40: lambda tag: decode_multi_dim_array(tag, column_major=False),
    64: lambda tag: decode_typed_array(tag, dtype="u1"),
    65: lambda tag: decode_typed_array(tag, dtype=">u2"),
    66: lambda tag: decode_typed_array(tag, dtype=">u4"),
    67: lambda tag: decode_typed_array(tag, dtype=">u8"),
    68: lambda tag: decode_typed_array(tag, dtype="u1"),
    69: lambda tag: decode_typed_array(tag, dtype="<u2"),
    70: lambda tag: decode_typed_array(tag, dtype="<u4"),
    71: lambda tag: decode_typed_array(tag, dtype="<u8"),
    72: lambda tag: decode_typed_array(tag, dtype="i1"),
    73: lambda tag: decode_typed_array(tag, dtype=">i2"),
    74: lambda tag: decode_typed_array(tag, dtype=">i4"),
    75: lambda tag: decode_typed_array(tag, dtype=">i8"),
    77: lambda tag: decode_typed_array(tag, dtype="<i2"),
    78: lambda tag: decode_typed_array(tag, dtype="<i4"),
    79: lambda tag: decode_typed_array(tag, dtype="<i8"),
    80: lambda tag: decode_typed_array(tag, dtype=">f2"),
    81: lambda tag: decode_typed_array(tag, dtype=">f4"),
    82: lambda tag: decode_typed_array(tag, dtype=">f8"),
    83: lambda tag: decode_typed_array(tag, dtype=">f16"),
    84: lambda tag: decode_typed_array(tag, dtype="<f2"),
    85: lambda tag: decode_typed_array(tag, dtype="<f4"),
    86: lambda tag: decode_typed_array(tag, dtype="<f8"),
    87: lambda tag: decode_typed_array(tag, dtype="<f16"),
    1040: lambda tag: decode_multi_dim_array(tag, column_major=True),
    56500: lambda tag: decode_dectris_compression(tag),
}


def stream2_tag_decoder(_decoder, tag):
    tag_decoder = tag_decoders.get(tag.tag)
    return tag_decoder(tag) if tag_decoder else tag



import logging
from collections.abc import Mapping
from dataclasses import dataclass, field, fields
from enum import Enum
from typing import Any

from .eiger_schema import (
    ro_float,
    ro_str,
    ro_uint,
    rw_bool,
    rw_float,
    rw_float_grid,
    rw_str,
    rw_uint,
    rw_uint_grid,
)

LOGGER = logging.getLogger(__name__)


FRAME_WIDTH: int = 4148
FRAME_HEIGHT: int = 4362


def config_keys() -> list[str]:
    return [
        "auto_summation",
        "beam_center_x",
        "beam_center_y",
        "bit_depth_image",
        "bit_depth_readout",
        "chi_increment",
        "chi_start",
        "compression",
        "count_time",
        "counting_mode",
        "countrate_correction_applied",
        "countrate_correction_count_cutoff",
        "data_collection_date",
        "description",
        "detector_distance",
        "detector_number",
        "detector_readout_time",
        "eiger_fw_version",
        "element",
        "extg_mode",
        "fast_arm",
        "flatfield_correction_applied",
        "frame_count_time",
        "frame_time",
        "incident_energy",
        "incident_particle_type",
        "instrument_name",
        "kappa_increment",
        "kappa_start",
        "mask_to_zero",
        "nexpi",
        "nimages",
        "ntrigger",
        "ntriggers_skipped",
        "number_of_excluded_pixels",
        "omega_increment",
        "omega_start",
        "phi_increment",
        "phi_start",
        "photon_energy",
        "pixel_mask_applied",
        "roi_mode",
        "sample_name",
        "sensor_material",
        "sensor_thickness",
        "software_version",
        "source_name",
        "threshold/1/energy",
        "threshold/1/mode",
        "threshold/1/number_of_excluded_pixels",
        "threshold/2/energy",
        "threshold/2/mode",
        "threshold/2/number_of_excluded_pixels",
        "threshold/difference/lower_threshold",
        "threshold/difference/mode",
        "threshold/difference/upper_threshold",
        "threshold_energy",
        "total_flux",
        "trigger_mode",
        "trigger_start_delay",
        "two_theta_increment",
        "two_theta_start",
        "virtual_pixel_correction_applied",
        # "wavelength",  # Eiger does not report wavelength as a key
        "x_pixel_size",
        "x_pixels_in_detector",
        "y_pixel_size",
        "y_pixels_in_detector",
    ]


class KAEnergy(Enum):
    """Possible element K-alpha energies for samples."""

    Li = 54.3
    Be = 108.5
    B = 183.3
    C = 277.0
    N = 392.4
    O = 524.9  # noqa: E741
    F = 676.8
    Ne = 848.6
    Na = 1040.98
    Mg = 1253.6
    Al = 1486.7
    Si = 1739.98
    P = 2013.7
    S = 2307.84
    Cl = 2622.39
    Ar = 2957.7
    K = 3313.8
    Ca = 3691.68
    Sc = 4090.6
    Ti = 4510.84
    V = 4952.2
    Cr = 5414.72
    Mn = 5898.75
    Fe = 6403.84
    Co = 6930.32
    Ni = 7478.15
    Cu = 8047.78
    Zn = 8638.86


@dataclass
class Threshold:
    """Data container for a single threshold configuration."""

    energy: float = field(default=6729.0, metadata=rw_float())
    mode: str = field(
        default="enabled", metadata=rw_str(allowed_values=["enabled", "disabled"])
    )
    number_of_excluded_pixels: int = field(default=0, metadata=ro_uint())

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        for field_ in fields(self):
            if field_.name == key:
                return {"value": vars(self)[field_.name], "metadata": field_.metadata}
        raise ValueError(f"No field with name {key}")

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: D105
        self.__dict__[key] = value


@dataclass
class ThresholdDifference:
    """Configuration for the threshold difference."""

    lower_threshold: int = field(default=1, metadata=ro_uint())
    mode: str = field(
        default="disabled", metadata=rw_str(allowed_values=["enabled", "disabled"])
    )
    upper_threshold: int = field(default=2, metadata=ro_uint())
    number_of_excluded_pixels: int = field(default=0, metadata=ro_uint())

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        for field_ in fields(self):
            if field_.name == key:
                return {"value": vars(self)[field_.name], "metadata": field_.metadata}
        raise ValueError(f"No field with name {key}")

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: D105
        self.__dict__[key] = value


def detector_translation() -> list[float]:
    return [0.0, 0.0, 0.0]


@dataclass
class EigerSettings:
    """A data container for Eiger device configuration."""

    auto_summation: bool = field(default=True, metadata=rw_bool())
    beam_center_x: float = field(default=0.0, metadata=rw_float())
    beam_center_y: float = field(default=0.0, metadata=rw_float())
    bit_depth_image: int = field(default=16, metadata=ro_uint())
    bit_depth_readout: int = field(default=16, metadata=ro_uint())
    chi_increment: float = field(default=0.0, metadata=rw_float())
    chi_start: float = field(default=0.0, metadata=rw_float())
    compression: str = field(
        default="bslz4", metadata=rw_str(allowed_values=["lz4", "bslz4", "none"])
    )
    count_time: float = field(default=0.1, metadata=rw_float())
    counting_mode: str = field(
        default="normal", metadata=rw_str(allowed_values=["normal", "retrigger"])
    )
    countrate_correction_applied: bool = field(default=True, metadata=rw_bool())
    countrate_correction_count_cutoff: int = field(default=1000, metadata=ro_uint())
    data_collection_date: str = field(
        default="2021-30-09T16:30:00.000-01:00", metadata=ro_str()
    )
    description: str = field(
        default="Simulated Eiger X 16M Detector", metadata=ro_str()
    )
    detector_distance: float = field(default=2.0, metadata=rw_float())
    detector_number: str = field(default="EIGERSIM001", metadata=ro_str())
    detector_readout_time: float = field(default=0.01, metadata=ro_float())
    detector_translation: list[float] = field(
        default_factory=detector_translation, metadata=ro_float()
    )
    eiger_fw_version: str = field(default="1.8.0", metadata=ro_str())
    element: str = field(
        default="Co", metadata=rw_str(allowed_values=[*(e.name for e in KAEnergy)])
    )
    extg_mode: str = field(
        default="double", metadata=rw_str(allowed_values=["single", "double"])
    )
    fast_arm: bool = field(default=False, metadata=rw_bool())
    flatfield: list[list[float]] = field(
        default_factory=lambda: [[]], metadata=rw_float_grid()
    )
    flatfield_correction_applied: bool = field(default=True, metadata=rw_bool())
    frame_count_time: float = field(default=0.01, metadata=ro_float())
    frame_time: float = field(default=0.12, metadata=rw_float())
    frame_period: float = field(default=0.12, metadata=rw_float())
    incident_energy: float = field(default=13458.0, metadata=rw_float())
    incident_particle_type: str = field(default="photons", metadata=ro_str())
    instrument_name: str = field(default="", metadata=rw_str())
    kappa_increment: float = field(default=0.0, metadata=rw_float())
    kappa_start: float = field(default=0.0, metadata=rw_float())
    mask_to_zero: bool = field(default=False, metadata=rw_bool())
    nexpi: int = field(default=1, metadata=rw_uint())
    nimages: int = field(default=1, metadata=rw_uint())
    ntrigger: int = field(default=1, metadata=rw_uint())
    ntriggers_skipped: int = field(default=0, metadata=rw_uint())
    number_of_excluded_pixels: int = field(default=0, metadata=ro_uint())
    omega_increment: float = field(default=0.0, metadata=rw_float())
    omega_start: float = field(default=0.0, metadata=rw_float())
    phi_increment: float = field(default=0.0, metadata=rw_float())
    phi_start: float = field(default=0.0, metadata=rw_float())
    photon_energy: float = field(default=6930.32, metadata=rw_float())
    pixel_mask: list[list[int]] = field(
        default_factory=lambda: [[]], metadata=rw_uint_grid()
    )
    pixel_mask_applied: bool = field(default=False, metadata=rw_bool())
    roi_mode: str = field(
        default="disabled", metadata=rw_str(allowed_values=["disabled", "4M-L", "4M-R"])
    )
    sample_name: str = field(default="", metadata=rw_str())
    sensor_material: str = field(default="Silicon", metadata=ro_str())
    sensor_thickness: float = field(default=0.01, metadata=ro_float())
    software_version: str = field(default="0.1.0", metadata=ro_str())
    source_name: str = field(default="", metadata=rw_str())
    threshold_energy: float = field(default=4020.5, metadata=rw_float())
    total_flux: float = field(default=0.0, metadata=rw_float())
    trigger_mode: str = field(
        default="exts",
        metadata=rw_str(
            allowed_values=["eies", "exte", "extg", "exts", "inte", "ints"]
        ),
    )
    trigger_start_delay: float = field(default=0.0, metadata=rw_float(min=0.0))
    two_theta_increment: float = field(default=0.0, metadata=rw_float())
    two_theta_start: float = field(default=0.0, metadata=rw_float())
    virtual_pixel_correction_applied: bool = field(default=True, metadata=rw_bool())
    wavelength: float = field(default=1.0, metadata=rw_float())
    x_pixel_size: float = field(default=0.01, metadata=ro_float())
    x_pixels_in_detector: int = field(default=FRAME_WIDTH, metadata=ro_uint())
    y_pixel_size: float = field(default=0.01, metadata=ro_float())
    y_pixels_in_detector: int = field(default=FRAME_HEIGHT, metadata=ro_uint())

    keys: list[str] = field(default_factory=config_keys)

    def __post_init__(self):
        self._threshold_config = {
            "1": Threshold(),
            "2": Threshold(energy=18841.0),
            "difference": ThresholdDifference(),
        }

    @property
    def threshold_config(self):
        return self._threshold_config

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        for field_ in fields(self):
            if field_.name == key:
                return {"value": vars(self)[field_.name], "metadata": field_.metadata}
        raise ValueError(f"No field with name {key}")

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: D105
        self.__dict__[key] = value

        self._check_dependencies(key, value)

    def _check_dependencies(self, key, value):
        if key == "element":
            self.photon_energy = getattr(KAEnergy, value).value
            self.wavelength = (1240 / self.photon_energy) / 10  # to convert to Angstrom
            self._calc_threshold_energy()

        elif key == "photon_energy":
            self.element = ""

            hc = 1240
            self.wavelength = (hc / self.photon_energy) / 10  # to convert to Angstrom

            self._calc_threshold_energy()

        elif key == "wavelength":
            self.element = ""

            hc = 1240
            self.photon_energy = hc / (self.wavelength * 10)  # to convert from Angstrom

            self._calc_threshold_energy()

        elif key == "count_time":
            self.frame_time = self.count_time + self.detector_readout_time

    def _calc_threshold_energy(self):
        self.threshold_energy = 0.5 * self.photon_energy

        LOGGER.warning("Flatfield not recalculated.")

    def filtered(self, exclude_fields: list[str]) -> Mapping[str, Any]:
        return {
            fld.name: vars(self)[fld.name]
            for fld in fields(self)
            if fld not in exclude_fields
        }



import struct
from dataclasses import dataclass


@dataclass
class Status:
    """A regular Cryostream status packet."""

    length: int
    type_status: int
    gas_set_point: int
    gas_temp: int
    gas_error: int
    run_mode: int
    phase_id: int
    ramp_rate: int
    target_temp: int
    evap_temp: int
    suct_temp: int
    remaining: int
    gas_flow: int
    gas_heat: int
    evap_heat: int
    suct_heat: int
    line_pressure: int
    alarm_code: int
    run_time: int
    controller_number: int
    software_version: int
    evap_adjust: int
    status_bytes_string: str = ">BBHHhBBHHHHHBBBBBBHHBB"

    @classmethod
    def from_packed(cls, b: bytes) -> "Status":
        """Create a status packet from its packed byte format."""
        return cls(*struct.unpack(cls.status_bytes_string, b))

    def pack(self) -> bytes:
        """Perform serialization of the status packet.

        Returns:
            bytes: A serialized status packet.
        """
        status_bytes = struct.pack(
            self.status_bytes_string,
            self.length,
            self.type_status,
            self.gas_set_point,
            self.gas_temp,
            self.gas_error,
            self.run_mode,
            self.phase_id,
            self.ramp_rate,
            self.target_temp,
            self.evap_temp,
            self.suct_temp,
            self.remaining,
            self.gas_flow,
            self.gas_heat,
            self.evap_heat,
            self.suct_heat,
            self.line_pressure,
            self.alarm_code,
            self.run_time,
            self.controller_number,
            self.software_version,
            self.evap_adjust,
        )

        return status_bytes


@dataclass
class ExtendedStatus:
    """An extended Cryostream status packet."""

    length: int
    type_status: int
    gas_set_point: int
    gas_temp: int
    gas_error: int
    run_mode: int
    phase_id: int
    ramp_rate: int
    target_temp: int
    evap_temp: int
    suct_temp: int
    remaining: int
    gas_flow: int
    gas_heat: int
    evap_heat: int
    suct_heat: int
    line_pressure: int
    alarm_code: int
    run_time: int
    controller_number: int
    software_version: int
    evap_adjust: int
    turbo_mode: int
    hardware_type: int
    shutter_state: int  # 800 series do not support CryoShutter
    shutter_time: int  # 800 series do not support CryoShutter
    avg_gas_heat: int
    avg_suct_heat: int
    time_to_fill: int
    total_hours: int
    extended_packet_string: str = ">BBHHhBBHHHHHBBBBBBHHBBBBBBBBHH"

    @classmethod
    def from_packed(cls, b: bytes) -> "ExtendedStatus":
        """Create a status packet from its packed byte format."""
        return cls(*struct.unpack(cls.extended_packet_string, b))

    def pack(self) -> bytes:
        """Perform serialization of the extended status packet.

        Returns:
            bytes: A serialized extended status packet.
        """
        extended_status_bytes = struct.pack(
            self.extended_packet_string,
            self.length,
            self.type_status,
            self.gas_set_point,
            self.gas_temp,
            self.gas_error,
            self.run_mode,
            self.phase_id,
            self.ramp_rate,
            self.target_temp,
            self.evap_temp,
            self.suct_temp,
            self.remaining,
            self.gas_flow,
            self.gas_heat,
            self.evap_heat,
            self.suct_heat,
            self.line_pressure,
            self.alarm_code,
            self.run_time,
            self.controller_number,
            self.software_version,
            self.evap_adjust,
            self.turbo_mode,
            self.hardware_type,
            self.shutter_state,
            self.shutter_time,
            self.avg_gas_heat,
            self.avg_suct_heat,
            self.time_to_fill,
            self.total_hours,
        )
        return extended_status_bytes



from dataclasses import dataclass, field, fields
from typing import Any

from tickit_devices.eiger.eiger_schema import rw_str


def stream_config_keys() -> list[str]:
    return ["format", "header_appendix", "header_detail", "image_appendix", "mode"]


LEGACY_STREAM = "legacy"
CBOR_STREAM = "cbor"


@dataclass
class StreamConfig:
    """Eiger stream configuration taken from the API spec."""

    mode: str = field(
        default="enabled", metadata=rw_str(allowed_values=["enabled", "disabled"])
    )
    format: str = field(
        default=LEGACY_STREAM,
        metadata=rw_str(allowed_values=[LEGACY_STREAM, CBOR_STREAM]),
    )
    header_detail: str = field(
        default="basic", metadata=rw_str(allowed_values=["none", "basic", "all"])
    )
    header_appendix: str = field(default="", metadata=rw_str())
    image_appendix: str = field(default="", metadata=rw_str())

    keys: list[str] = field(default_factory=stream_config_keys)

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        for field_ in fields(self):
            if field_.name == key:
                return {"value": vars(self)[field_.name], "metadata": field_.metadata}
        raise ValueError(f"No field with name {key}")

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: D105
        self.__dict__[key] = value



import re
from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Generic, TypeVar, get_type_hints

import pydantic
from tickit.core.components.component import ComponentConfig
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime


@dataclass
class Param:
    """
    Representation of the internal configuration of the Zebra.
    Sometimes is shared state between multiple blocks.
    """

    reg: int
    blocks: list[str]


@dataclass
class Mux:
    """
    Representation of the internal mapping of the Zebra.
    e.g If AND1_INP1=Mux(0x08, ...) is set to 0x09, then
    AND1_INP1 should take the same input as AND1_INP2=Mux(0x09, ...)

    """

    reg: int
    block: str | None = None


DIVS = [f"DIV{i}" for i in range(1, 5)]
PULSES = [f"PULSE{i}" for i in range(1, 5)]
GATES = [f"GATE{i}" for i in range(1, 5)]

register_types: dict[str, Param | Mux] = {
    "AND1_INV": Param(0x00, ["AND1"]),
    "AND2_INV": Param(0x01, ["AND2"]),
    "AND3_INV": Param(0x02, ["AND3"]),
    "AND4_INV": Param(0x03, ["AND4"]),
    "AND1_ENA": Param(0x04, ["AND1"]),
    "AND2_ENA": Param(0x05, ["AND2"]),
    "AND3_ENA": Param(0x06, ["AND3"]),
    "AND4_ENA": Param(0x07, ["AND4"]),
    "AND1_INP1": Mux(0x08, "AND1"),
    "AND1_INP2": Mux(0x09, "AND1"),
    "AND1_INP3": Mux(0x0A, "AND1"),
    "AND1_INP4": Mux(0x0B, "AND1"),
    "AND2_INP1": Mux(0x0C, "AND2"),
    "AND2_INP2": Mux(0x0D, "AND2"),
    "AND2_INP3": Mux(0x0E, "AND2"),
    "AND2_INP4": Mux(0x0F, "AND2"),
    "AND3_INP1": Mux(0x10, "AND3"),
    "AND3_INP2": Mux(0x11, "AND3"),
    "AND3_INP3": Mux(0x12, "AND3"),
    "AND3_INP4": Mux(0x13, "AND3"),
    "AND4_INP1": Mux(0x14, "AND4"),
    "AND4_INP2": Mux(0x15, "AND4"),
    "AND4_INP3": Mux(0x16, "AND4"),
    "AND4_INP4": Mux(0x17, "AND4"),
    "OR1_INV": Param(0x18, ["OR1"]),
    "OR2_INV": Param(0x19, ["OR2"]),
    "OR3_INV": Param(0x1A, ["OR3"]),
    "OR4_INV": Param(0x1B, ["OR4"]),
    "OR1_ENA": Param(0x1C, ["OR1"]),
    "OR2_ENA": Param(0x1D, ["OR2"]),
    "OR3_ENA": Param(0x1E, ["OR3"]),
    "OR4_ENA": Param(0x1F, ["OR4"]),
    "OR1_INP1": Mux(0x20, "OR1"),
    "OR1_INP2": Mux(0x21, "OR1"),
    "OR1_INP3": Mux(0x22, "OR1"),
    "OR1_INP4": Mux(0x23, "OR1"),
    "OR2_INP1": Mux(0x24, "OR2"),
    "OR2_INP2": Mux(0x25, "OR2"),
    "OR2_INP3": Mux(0x26, "OR2"),
    "OR2_INP4": Mux(0x27, "OR2"),
    "OR3_INP1": Mux(0x28, "OR3"),
    "OR3_INP2": Mux(0x29, "OR3"),
    "OR3_INP3": Mux(0x2A, "OR3"),
    "OR3_INP4": Mux(0x2B, "OR3"),
    "OR4_INP1": Mux(0x2C, "OR4"),
    "OR4_INP2": Mux(0x2D, "OR4"),
    "OR4_INP3": Mux(0x2E, "OR4"),
    "OR4_INP4": Mux(0x2F, "OR4"),
    "GATE1_INP1": Mux(0x30, "GATE1"),
    "GATE2_INP1": Mux(0x31, "GATE2"),
    "GATE3_INP1": Mux(0x32, "GATE3"),
    "GATE4_INP1": Mux(0x33, "GATE4"),
    "GATE1_INP2": Mux(0x34, "GATE1"),
    "GATE2_INP2": Mux(0x35, "GATE2"),
    "GATE3_INP2": Mux(0x36, "GATE3"),
    "GATE4_INP2": Mux(0x37, "GATE4"),
    "DIV1_DIVLO": Param(0x38, ["DIV1"]),
    "DIV1_DIVHI": Param(0x39, ["DIV1"]),
    "DIV2_DIVLO": Param(0x3A, ["DIV2"]),
    "DIV2_DIVHI": Param(0x3B, ["DIV2"]),
    "DIV3_DIVLO": Param(0x3C, ["DIV3"]),
    "DIV3_DIVHI": Param(0x3D, ["DIV3"]),
    "DIV4_DIVLO": Param(0x3E, ["DIV4"]),
    "DIV4_DIVHI": Param(0x3F, ["DIV4"]),
    "DIV1_INP": Mux(0x40, "DIV1"),
    "DIV2_INP": Mux(0x41, "DIV2"),
    "DIV3_INP": Mux(0x42, "DIV3"),
    "DIV4_INP": Mux(0x43, "DIV4"),
    "PULSE1_DLY": Param(0x44, ["PULSE1"]),
    "PULSE2_DLY": Param(0x45, ["PULSE2"]),
    "PULSE3_DLY": Param(0x46, ["PULSE3"]),
    "PULSE4_DLY": Param(0x47, ["PULSE4"]),
    "PULSE1_WID": Param(0x48, ["PULSE1"]),
    "PULSE2_WID": Param(0x49, ["PULSE2"]),
    "PULSE3_WID": Param(0x4A, ["PULSE3"]),
    "PULSE4_WID": Param(0x4B, ["PULSE4"]),
    "PULSE1_PRE": Param(0x4C, ["PULSE1"]),
    "PULSE2_PRE": Param(0x4D, ["PULSE2"]),
    "PULSE3_PRE": Param(0x4E, ["PULSE3"]),
    "PULSE4_PRE": Param(0x4F, ["PULSE4"]),
    "PULSE1_INP": Mux(0x50, "PULSE1"),
    "PULSE2_INP": Mux(0x51, "PULSE2"),
    "PULSE3_INP": Mux(0x52, "PULSE3"),
    "PULSE4_INP": Mux(0x53, "PULSE4"),
    "POLARITY": Param(0x54, GATES + DIVS + PULSES),
    "QUAD_DIR": Param(0x55, ["QUAD"]),
    "QUAD_STEP": Param(0x56, ["QUAD"]),
    "PC_ARM_INP": Mux(0x57, "PC"),
    "PC_GATE_INP": Mux(0x58, "PC"),
    "PC_PULSE_INP": Mux(0x59, "PC"),
    "OUT1_TTL": Mux(0x60),
    "OUT1_NIM": Mux(0x61),
    "OUT1_LVDS": Mux(0x62),
    "OUT2_TTL": Mux(0x63),
    "OUT2_NIM": Mux(0x64),
    "OUT2_LVDS": Mux(0x65),
    "OUT3_TTL": Mux(0x66),
    "OUT3_OC": Mux(0x67),
    "OUT3_LVDS": Mux(0x68),
    "OUT4_TTL": Mux(0x69),
    "OUT4_NIM": Mux(0x6A),
    "OUT4_PECL": Mux(0x6B),
    "OUT5_ENCA": Mux(0x6C),
    "OUT5_ENCB": Mux(0x6D),
    "OUT5_ENCZ": Mux(0x6E),
    "OUT5_CONN": Mux(0x6F),
    "OUT6_ENCA": Mux(0x70),
    "OUT6_ENCB": Mux(0x71),
    "OUT6_ENCZ": Mux(0x72),
    "OUT6_CONN": Mux(0x73),
    "OUT7_ENCA": Mux(0x74),
    "OUT7_ENCB": Mux(0x75),
    "OUT7_ENCZ": Mux(0x76),
    "OUT7_CONN": Mux(0x77),
    "OUT8_ENCA": Mux(0x78),
    "OUT8_ENCB": Mux(0x79),
    "OUT8_ENCZ": Mux(0x7A),
    "OUT8_CONN": Mux(0x7B),
    "DIV_FIRST": Param(0x7C, DIVS),
    "SYS_RESET": Param(0x7E, []),
    "SOFT_IN": Param(0x7F, ["SOFT"]),
    "POS1_SETLO": Param(0x80, ["POS1"]),
    "POS1_SETHI": Param(0x81, ["POS1"]),
    "POS2_SETLO": Param(0x82, ["POS2"]),
    "POS2_SETHI": Param(0x83, ["POS2"]),
    "POS3_SETLO": Param(0x84, ["POS3"]),
    "POS3_SETHI": Param(0x85, ["POS3"]),
    "POS4_SETLO": Param(0x86, ["POS4"]),
    "POS4_SETHI": Param(0x87, ["POS4"]),
    "PC_ENC": Param(0x88, ["PC"]),
    "PC_TSPRE": Param(0x89, ["PC"]),
    "PC_ARM_SEL": Param(0x8A, ["PC"]),
    "PC_ARM": Param(0x8B, ["PC"]),
    "PC_DISARM": Param(0x8C, ["PC"]),
    "PC_GATE_SEL": Param(0x8D, ["PC"]),
    "PC_GATE_STARTLO": Param(0x8E, ["PC"]),
    "PC_GATE_STARTHI": Param(0x8F, ["PC"]),
    "PC_GATE_WIDLO": Param(0x90, ["PC"]),
    "PC_GATE_WIDHI": Param(0x91, ["PC"]),
    "PC_GATE_NGATELO": Param(0x92, ["PC"]),
    "PC_GATE_NGATEHI": Param(0x93, ["PC"]),
    "PC_GATE_STEPLO": Param(0x94, ["PC"]),
    "PC_GATE_STEPHI": Param(0x95, ["PC"]),
    "PC_PULSE_SEL": Param(0x96, ["PC"]),
    "PC_PULSE_STARTLO": Param(0x97, ["PC"]),
    "PC_PULSE_STARTHI": Param(0x98, ["PC"]),
    "PC_PULSE_WIDLO": Param(0x99, ["PC"]),
    "PC_PULSE_WIDHI": Param(0x9A, ["PC"]),
    "PC_PULSE_STEPLO": Param(0x9B, ["PC"]),
    "PC_PULSE_STEPHI": Param(0x9C, ["PC"]),
    "PC_PULSE_MAXLO": Param(0x9D, ["PC"]),
    "PC_PULSE_MAXHI": Param(0x9E, ["PC"]),
    "PC_BIT_CAP": Param(0x9F, ["PC"]),
    "PC_DIR": Param(0xA0, ["PC"]),
    "PC_PULSE_DLYLO": Param(0xA1, ["PC"]),
    "PC_PULSE_DLYHI": Param(0xA2, ["PC"]),
    "SYS_VER": Param(0xF0, []),
    "SYS_STATERR": Param(0xF1, []),
    "SYS_STAT1LO": Param(0xF2, []),
    "SYS_STAT1HI": Param(0xF3, []),
    "SYS_STAT2LO": Param(0xF4, []),
    "SYS_STAT2HI": Param(0xF5, []),
    "PC_NUM_CAPLO": Param(0xF6, []),
    "PC_NUM_CAPHI": Param(0xF7, []),
}

register_names = {reg.reg: name for name, reg in register_types.items()}
param_types = {name: t for name, t in register_types.items() if isinstance(t, Param)}
mux_types = {name: t for name, t in register_types.items() if isinstance(t, Mux)}

Inputs = TypeVar("Inputs")
Outputs = TypeVar("Outputs")


@dataclass
class Block(Device, ABC, Generic[Inputs, Outputs]):
    name: str
    """
    Blocks take 20ns to propagate a signal across, so we cache the next output value
    and emit it 20ns later.
    """
    previous_outputs: Outputs
    next_outputs: Outputs | None = None
    params: dict[str, int] | None = None

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        if self.next_outputs:
            self.previous_outputs = self.next_outputs
            self.next_outputs = None
            return DeviceUpdate(self.previous_outputs, call_at=None)
        else:
            self.next_outputs = self._get_next_outputs(inputs)
            return DeviceUpdate(self.previous_outputs, call_at=SimTime(time + 20))

    @abstractmethod
    def _get_next_outputs(self, inputs: Inputs) -> Outputs: ...

    @property
    def num(self):
        match = re.search(r"\d*$", self.name)
        assert match, f"No trailing number in {self.name}"
        return int(match.group())

    def read_mux(self, register: str) -> int:
        return 0

    def set_mux(self, register: str, value: int) -> int:
        return 0


@pydantic.v1.dataclasses.dataclass
class BlockConfig(ComponentConfig, ABC):
    @abstractmethod
    def __call__(self) -> Block:
        """Create a block"""
        ...


def default_filler(typed_dict_type) -> Callable[[], Any]:
    def make_default():
        return {name: typ() for name, typ in get_type_hints(typed_dict_type).items()}

    return make_default


def extract_bit(registers: dict[str, int], key: str, shift: int) -> bool:
    return bool((registers[key] >> shift) & 1)


def clear_bit(registers: dict[str, int], key: str, shift: int):
    registers[key] &= ~(1 << shift)


def set_bit(registers: dict[str, int], key: str, shift: int):
    registers[key] |= 1 << shift


def rising(old: bool, new: bool) -> bool:
    return new and not old


def in_ns(ticks: int | None) -> SimTime | None:
    return None if ticks is None else SimTime(ticks * 20)



import logging
from collections.abc import Iterable, Mapping
from queue import Queue
from typing import Any, TypedDict

import numpy as np
from pydantic.v1 import BaseModel
from tickit.core.typedefs import SimTime

from tickit_devices.eiger.data.dummy_image import Image
from tickit_devices.eiger.data.schema import (
    AcquisitionDetailsHeader,
    AcquisitionSeriesFooter,
    AcquisitionSeriesHeader,
    ImageCharacteristicsHeader,
    ImageConfigHeader,
    ImageHeader,
)
from tickit_devices.eiger.eiger_settings import EigerSettings

LOGGER = logging.getLogger(__name__)


_Message = BaseModel | Mapping[str, Any] | bytes


class EigerStream:
    """Simulation of an Eiger stream."""

    callback_period: SimTime

    _message_buffer: Queue[_Message]

    class Inputs(TypedDict): ...

    class Outputs(TypedDict): ...

    def __init__(self, callback_period: int = int(1e9)) -> None:
        """An Eiger Stream constructor."""
        self.callback_period = SimTime(callback_period)

        self._message_buffer = Queue()

    def begin_series(
        self, settings: EigerSettings, series_id: int, header_detail: str
    ) -> None:
        """Send the headers marking the beginning of the acquisition series.

        Args:
            settings: Current detector configuration, a snapshot may be sent with the
                headers.
            series_id: ID for the acquisition series.
            header_detail: Header detail for start message - "none", "basic" or "all"
        """
        header = AcquisitionSeriesHeader(
            header_detail=header_detail,
            series=series_id,
        )
        self._buffer(header)

        if header_detail != "none":
            config_header = settings.filtered(
                ["flatfield", "pixelmask", "countrate_correction_table"]
            )
            self._buffer(config_header)

            if header_detail == "all":
                x = settings.x_pixels_in_detector
                y = settings.y_pixels_in_detector

                flatfield_header = AcquisitionDetailsHeader(
                    htype="flatfield-1.0",
                    shape=(x, y),
                    type="float32",
                )
                self._buffer(flatfield_header)
                self._buffer(np.zeros(shape=(y, x), dtype="float32").tobytes())

                pixel_mask_header = AcquisitionDetailsHeader(
                    htype="dpixelmask-1.0",
                    shape=(x, y),
                    type="uint32",
                )
                self._buffer(pixel_mask_header)
                self._buffer(np.zeros(shape=(y, x), dtype="uint32").tobytes())

                countrate_table_header = AcquisitionDetailsHeader(
                    htype="dcountrate_table-1.0",
                    shape=(2, 1000),
                    type="float32",
                )
                self._buffer(countrate_table_header)
                self._buffer(np.zeros(shape=(1000, 2), dtype="float32").tobytes())

    def insert_image(self, image: Image, series_id: int) -> None:
        """Send headers and an data blob for a single image.

        Args:
            image: The image with associated metadata
            series_id: ID for the acquisition series.
        """
        header = ImageHeader(
            frame=image.index,
            hash=image.hash,
            series=series_id,
        )
        characteristics_header = ImageCharacteristicsHeader(
            encoding=image.encoding,
            shape=image.shape,
            size=len(image.data),
            type=image.dtype,
        )
        config_header = ImageConfigHeader(
            real_time=0,
            start_time=0,
            stop_time=0,
        )

        self._buffer(header)
        self._buffer(characteristics_header)
        self._buffer(image.data)
        self._buffer(config_header)

    def end_series(self, series_id: int) -> None:
        """Send footer marking the end of an acquisition series.

        Args:
            series_id: ID of the series to end.
        """
        footer = AcquisitionSeriesFooter(series=series_id)
        self._buffer(footer)

    def consume_data(self) -> Iterable[_Message]:
        """Consume all headers and data buffered by other methods.

        Returns:
            Iterable[_Message]: Iterable of headers and data
        """
        while not self._message_buffer.empty():
            yield self._message_buffer.get()

    def _buffer(self, message: _Message) -> None:
        self._message_buffer.put_nowait(message)



import logging
from ctypes import c_short, c_ubyte, c_ushort

from tickit_devices.cryostream.states import AlarmCodes, PhaseIds, RunModes
from tickit_devices.cryostream.status import ExtendedStatus, Status

LOGGER = logging.getLogger(__name__)


class CryostreamBase:
    """A base class for cryostream device logic."""

    min_temp: int = 8000  # cK
    max_temp: int = 40000  # ck
    min_rate: int = 1
    max_rate: int = 360
    min_plat_duration: int = 1
    max_plat_duration: int = 1440
    default_temp_shutdown: int = 30000
    default_ramp_rate: int = 360
    random_value: int = 20  # for status packet unknown values

    def __init__(self) -> None:
        """A CryostreamBase constructor which assigns initial values."""
        self.run_mode: int = 0
        self.phase_id: int = 0
        self.alarm_code: int = 0
        self.turbo_mode: int = 0
        self.gas_temp: int = 30000
        self.ramp_rate: int = 0
        self.gas_flow: int = 0
        self._target_temp: int = 0
        self.time_at_last_update: float = 0.0

    async def restart(self) -> None:
        """Stop Cryostream and re-initialise system back to Ready."""
        start_target_temp = 10000
        await self.stop()
        self.run_mode = RunModes.STARTUP.value
        await self.cool(start_target_temp)
        if self.gas_temp == start_target_temp:
            self.run_mode = RunModes.STARTUPOK.value
        else:
            self.run_mode = RunModes.STARTUPFAIL.value

    async def ramp(self, ramp_rate: int, target_temp: int) -> None:
        """Change gas temperature to a set value at a controlled rate.

        Args:
            ramp_rate (int): The rate at which the temperature should change.
            target_temp (int): The target temperature.
        """
        if ramp_rate < self.min_rate or ramp_rate > self.max_rate:
            self.alarm_code = AlarmCodes.TEMP_CONTROL_ERROR
        else:
            self.ramp_rate = ramp_rate

        if target_temp < self.min_temp or target_temp > self.max_temp:
            self.alarm_code = AlarmCodes.TEMP_CONTROL_ERROR
            # note that the target temperature is not changed.
        else:
            self._target_temp = target_temp

        # Todo
        """
        difference_temp = abs(self.gas_temp - self.target_temp)
        if 10 < difference_temp < 25:
            self.alarm_code = AlarmCodes.TEMP_WARNING
        if difference_temp > 25:
            self.alarm_code = AlarmCodes.TEMP_CONTROL_ERROR
        """

        self.run_mode = RunModes.RUN.value
        if self.phase_id != PhaseIds.COOL.value:
            self.phase_id = PhaseIds.RAMP.value

        if (
            self.phase_id == PhaseIds.COOL.value
            or (self.turbo_mode == 1 and self._target_temp < 31000)
            or self._target_temp < 9000
        ):
            self.gas_flow = 10
        else:
            self.gas_flow = 5

    async def plat(self, duration: int) -> None:
        """Maintain the current temperature for a set amount of time.

        Args:
            duration (int): The duration for which the temperature should be held.
        """
        if duration < self.min_plat_duration:
            raise ValueError("Duration set to less than minimum plat duration.")
        if duration > self.max_plat_duration:
            raise ValueError("Duration set to more than maximum plat duration.")

        self.plat_duration = duration
        self._target_temp = self.gas_temp
        self.run_mode = RunModes.RUN.value
        self.phase_id = PhaseIds.PLAT.value

    async def hold(self) -> None:
        """Maintain the current temperature indefinitely."""
        self.run_mode = RunModes.RUN.value
        self.phase_id = PhaseIds.HOLD.value
        self._target_temp = self.gas_temp

    async def cool(self, target_temp: int) -> None:
        """Make gas temperature decrease to a set value as quickly as possible.

        Args:
            target_temp (int): The target temperature.
        """
        self.run_mode = RunModes.RUN
        self.phase_id = PhaseIds.COOL.value
        self._target_temp = target_temp
        await self.ramp(self.default_ramp_rate, self._target_temp)

    async def end(self, ramp_rate: int) -> None:
        """Bring the gas temperature to 300 K at ramp rate, then halt and stop.

        Args:
            ramp_rate (int): The rate at which the temperature should change.
        """
        if self.run_mode not in (
            RunModes.SHUTDOWNOK.value,
            RunModes.SHUTDOWNFAIL.value,
        ):
            self.phase_id = PhaseIds.END.value
            await self.ramp(ramp_rate, self.default_temp_shutdown)
            if self.gas_temp == self.default_temp_shutdown:
                self.gas_flow = 0
                self.run_mode = RunModes.SHUTDOWNOK.value
            else:
                self.run_mode = RunModes.SHUTDOWNFAIL.value

    async def purge(self) -> None:
        """Bring the gas temperature to 300 K at max rate, then halt and stop."""
        if self.run_mode not in (
            RunModes.SHUTDOWNOK.value,
            RunModes.SHUTDOWNFAIL.value,
        ):
            self.phase_id = PhaseIds.PURGE.value
            self.gas_flow = 0
            await self.ramp(self.default_ramp_rate, self.default_temp_shutdown)
            self.phase_id = PhaseIds.PURGE.value
            if self.gas_temp == self.default_temp_shutdown:
                self.run_mode = RunModes.SHUTDOWNOK.value
            else:
                self.run_mode = RunModes.SHUTDOWNFAIL.value

    async def pause(self) -> None:
        """Interrupt and maintain the current gas temperature until resumed."""
        # TODO interrupt other commands from running
        ...

    async def resume(self) -> None:
        """Resume the previous command."""
        if self.phase_id == PhaseIds.HOLD.value:
            LOGGER.warn("Cannot return to previous command")  # TODO keep commands

    async def stop(self) -> None:
        """Gas flow is halted and the system is stopped at the current temperature."""
        if self.run_mode not in (
            RunModes.SHUTDOWNOK.value,
            RunModes.SHUTDOWNFAIL.value,
        ):
            self.gas_flow = 0
            self._target_temp = self.gas_temp
            self.run_mode = RunModes.SHUTDOWNOK.value

    async def turbo(self, turbo_on: int) -> None:
        """Set turbo mode to use maximum achievable flow.

        The system will use 10 l/min gas flow except above 310 K
        Above 310 K available heater power limits the maximum achievable
        flow to 5 l/min.

        Args:
            turbo_on (int): The desired turbo mode, where 0 denotes off and 1 denotes
                on.
        """
        if turbo_on == 1:
            self.turbo_mode = 1
            if self.gas_temp < 310:
                self.gas_flow = 10
            else:
                self.gas_flow = 5
        else:
            self.turbo_mode = 0

    def update_temperature(self, time: float) -> int:
        """Update the Cryostream gas temperature according to mode and time.

        Args:
            time (float): The current simulation time (in nanoseconds).

        Returns:
            int: The current gas temperature.
        """
        delta_time = (time - self.time_at_last_update) / 1e9
        difference = self._target_temp - self.gas_temp
        if abs(self.gas_temp - self._target_temp) < 10:
            self.phase_id = PhaseIds.HOLD.value
        delta_temp = self.gas_flow * delta_time
        if difference < 0:
            delta_temp = -delta_temp
        self.gas_temp = int(self.gas_temp + delta_temp)
        self.time_at_last_update = time

        return self.gas_temp

    async def set_status_format(self, status_format: int) -> None:
        """Sets the status packet format.

        Args:
            status_format (int): The status packet format, where 0 denotes a standard
                status packet and 1 denotes an extended status packet.
        """
        if status_format == 0:
            self.status = Status(
                length=c_ubyte(32).value,  # 0R
                type_status=c_ubyte(1).value,  # 0R
                gas_set_point=c_ushort(22).value,  # 02R
                gas_temp=c_ushort(22).value,  # 02r
                gas_error=c_short(22).value,
                run_mode=c_ubyte(self.run_mode).value,
                phase_id=c_ubyte(self.phase_id).value,
                ramp_rate=c_short(self.ramp_rate).value,
                target_temp=c_ushort(self._target_temp).value,
                evap_temp=c_ushort(22).value,
                suct_temp=c_ushort(22).value,
                remaining=c_ushort(88).value,
                gas_flow=c_ubyte(self.gas_flow).value,
                gas_heat=c_ubyte(22).value,
                evap_heat=c_ubyte(22).value,
                suct_heat=c_ubyte(22).value,
                line_pressure=c_ubyte(10).value,
                alarm_code=c_ubyte(self.alarm_code).value,
                run_time=c_ushort(100).value,
                controller_number=c_ushort(10).value,
                software_version=c_ubyte(12).value,
                evap_adjust=c_ubyte(120).value,
            )

        if status_format == 1:
            self.extended_status = ExtendedStatus(
                length=c_ubyte(42).value,
                type_status=c_ubyte(2).value,
                gas_set_point=c_ushort(22).value,
                gas_temp=c_ushort(22).value,
                gas_error=c_short(22).value,
                run_mode=c_ubyte(self.run_mode).value,
                phase_id=c_ubyte(self.phase_id).value,
                ramp_rate=c_short(self.ramp_rate).value,
                target_temp=c_ushort(self._target_temp).value,
                evap_temp=c_ushort(22).value,
                suct_temp=c_ushort(22).value,
                remaining=c_ushort(88).value,
                gas_flow=c_ubyte(self.gas_flow).value,
                gas_heat=c_ubyte(22).value,
                evap_heat=c_ubyte(22).value,
                suct_heat=c_ubyte(22).value,
                line_pressure=c_ubyte(10).value,
                alarm_code=c_ubyte(self.alarm_code).value,
                run_time=c_ushort(100).value,
                controller_number=c_ushort(10).value,
                software_version=c_ubyte(12).value,
                evap_adjust=c_ubyte(120).value,
                turbo_mode=c_ubyte(self.turbo_mode).value,
                hardware_type=c_ubyte(1).value,
                shutter_state=c_ubyte(22).value,
                shutter_time=c_ubyte(22).value,
                avg_gas_heat=c_ubyte(22).value,
                avg_suct_heat=c_ubyte(22).value,
                time_to_fill=c_ushort(22).value,
                total_hours=c_ushort(22).value,
            )
            self.extended_status.gas_temp = c_ushort(self.gas_temp).value
            self.extended_status.run_mode = c_ubyte(self.run_mode).value
            self.extended_status.phase_id = c_ubyte(self.phase_id).value
            self.extended_status.target_temp = c_ushort(self._target_temp).value
            self.extended_status.ramp_rate = c_short(self.ramp_rate).value
            self.extended_status.gas_flow = c_ubyte(self.gas_flow).value
            self.extended_status.alarm_code = c_ubyte(self.alarm_code).value
            self.extended_status.turbo_mode = c_ubyte(self.turbo_mode).value

    async def get_status(self, status_format: int) -> Status | ExtendedStatus:
        """Get a Status or ExtendedStatus packet.

        Args:
            status_format (int): The status packet format, where 0 denotes a standard
                status packet and 1 denotes an extended status packet.

        Returns:
            Union[
                tickit.devices.cryostream.status.Status,
                tickit.devices.cryostream.status.ExtendedStatus]: The status packet.
        """
        if status_format == 0:
            if hasattr(self, "status"):
                pass
            else:
                await self.set_status_format(0)

            return self.status

        if status_format == 1:
            if hasattr(self, "extended_status"):
                pass
            else:
                await self.set_status_format(1)

            return self.extended_status

        else:
            raise ValueError("Invalid status format parameter.")



import logging

import pydantic.v1.dataclasses
from tickit.adapters.io import HttpIo, ZeroMqPushIo
from tickit.core.adapter import AdapterContainer
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_component import DeviceComponent

from tickit_devices.eiger.eiger import EigerDevice
from tickit_devices.eiger.eiger_adapters import EigerRESTAdapter, EigerZMQAdapter
from tickit_devices.eiger.stream.stream_config import CBOR_STREAM, LEGACY_STREAM


@pydantic.v1.dataclasses.dataclass
class Eiger(ComponentConfig):
    """Eiger simulation with HTTP adapter."""

    host: str = "0.0.0.0"
    port: int = 8081
    stream_host: str = "127.0.0.1"
    stream_legacy_port: int = 9999
    stream_cbor_port: int = 31001

    def __call__(self) -> Component:  # noqa: D102
        logging.getLogger("aiohttp.access").setLevel(logging.WARNING)
        device = EigerDevice()
        adapters = [
            AdapterContainer(
                EigerRESTAdapter(device),
                HttpIo(
                    self.host,
                    self.port,
                ),
            ),
            AdapterContainer(
                EigerZMQAdapter(device.streams[LEGACY_STREAM]),
                ZeroMqPushIo(
                    self.stream_host,
                    self.stream_legacy_port,
                ),
            ),
            AdapterContainer(
                EigerZMQAdapter(device.streams[CBOR_STREAM]),
                ZeroMqPushIo(
                    self.stream_host,
                    self.stream_cbor_port,
                ),
            ),
        ]
        return DeviceComponent(
            name=self.name,
            device=device,
            adapters=adapters,
        )



import pytest

from tickit_devices.eiger.eiger_settings import EigerSettings, KAEnergy

# # # # # EigerStatus Tests # # # # #


@pytest.fixture
def eiger_settings() -> EigerSettings:
    return EigerSettings()


def test_eiger_settings_constructor():
    EigerSettings()


def test_eiger_settings_get_set(eiger_settings):
    assert eiger_settings["count_time"]["value"] == 0.1
    eiger_settings["count_time"] = 0.2
    assert eiger_settings["count_time"]["value"] == 0.2

    with pytest.raises(ValueError):
        eiger_settings["doesnt_exist"]


def test_eiger_settings_get_element(eiger_settings):
    assert "Co" == eiger_settings.element


def test_eiger_settings_set_element(eiger_settings):
    eiger_settings["element"] = "Li"

    assert "Li" == eiger_settings.element
    assert KAEnergy["Li"].value == eiger_settings.photon_energy
    assert (1240 / eiger_settings.photon_energy) / 10 == eiger_settings.wavelength
    assert 0.5 * eiger_settings.photon_energy == eiger_settings.threshold_energy


def test_eiger_settings_set_photon_energy(eiger_settings):
    eiger_settings["photon_energy"] = 1000.0

    assert 1000.0 == eiger_settings.photon_energy
    assert "" == eiger_settings.element
    assert (1240 / eiger_settings.photon_energy) / 10 == eiger_settings.wavelength
    assert 0.5 * eiger_settings.photon_energy == eiger_settings.threshold_energy


def test_eiger_settings_set_wavelength(eiger_settings):
    eiger_settings["wavelength"] = 1.24

    assert 1.24 == eiger_settings.wavelength
    assert "" == eiger_settings.element
    assert 1240 / (eiger_settings.wavelength * 10) == eiger_settings.photon_energy
    assert 0.5 * eiger_settings.photon_energy == eiger_settings.threshold_energy


def test_eiger_settings_set_count_time(eiger_settings):
    eiger_settings["count_time"] = 0.2

    assert 0.2 == eiger_settings.count_time
    assert (
        eiger_settings.count_time + eiger_settings.detector_readout_time
        == eiger_settings.frame_time
    )


def test_eiger_settings_threshold_config(eiger_settings):
    assert eiger_settings.threshold_config["1"]["energy"]["value"] == 6729
    eiger_settings.threshold_config["1"]["energy"] = 6829
    assert eiger_settings.threshold_config["1"]["energy"]["value"] == 6829

    with pytest.raises(ValueError):
        eiger_settings.threshold_config["1"]["doesnt_exist"]

    assert eiger_settings.threshold_config["difference"]["mode"]["value"] == "disabled"
    eiger_settings.threshold_config["difference"]["mode"] = "enabled"
    assert eiger_settings.threshold_config["difference"]["mode"]["value"] == "enabled"

    with pytest.raises(ValueError):
        eiger_settings.threshold_config["difference"]["doesnt_exist"]
